# coding:utf-8
#
# unike/module/model/TransR.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 25, 2024
# 
# 该头文件定义了 TransR.

"""
TransR - 是一个为实体和关系嵌入向量分别构建了独立的向量空间，将实体向量投影到特定的关系向量空间进行平移操作的模型。
"""

import torch
import typing
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from .Model import Model
from typing_extensions import override

class TransR(Model):

	"""
	``TransR`` :cite:`TransR` 提出于 2015 年，是一个为实体和关系嵌入向量分别构建了独立的向量空间，将实体向量投影到特定的关系向量空间进行平移操作的模型。
	
	评分函数为:
	
	.. math::

		\Vert hM_r+r-tM_r \Vert_{L_1/L_2}

	正三元组的评分函数的值越小越好，如果想获得更详细的信息请访问 :ref:`TransR <transr>`。

	例子::

		from unike.data import KGEDataLoader, BernSampler, TradTestSampler
		from unike.module.model import TransE, TransR
		from unike.module.loss import MarginLoss
		from unike.module.strategy import NegativeSampling
		from unike.config import Trainer, Tester
		
		# dataloader for training
		dataloader = KGEDataLoader(
			in_path = "../../benchmarks/FB15K237/", 
			batch_size = 2048,
			neg_ent = 25, 
			test = True,
			test_batch_size = 10,
			num_workers = 16,
			train_sampler = BernSampler,
			test_sampler = TradTestSampler
		)
		
		# define the transe
		transe = TransE(
			ent_tol = dataloader.get_ent_tol(),
			rel_tol = dataloader.get_rel_tol(),
			dim = 100, 
			p_norm = 1, 
			norm_flag = True)
		
		transr = TransR(
			ent_tol = dataloader.get_ent_tol(),
			rel_tol = dataloader.get_rel_tol(),
			dim_e = 100,
			dim_r = 100,
			p_norm = 1, 
			norm_flag = True,
			rand_init = False)
		
		model_e = NegativeSampling(
			model = transe, 
			loss = MarginLoss(margin = 5.0)
		)
		
		model_r = NegativeSampling(
			model = transr,
			loss = MarginLoss(margin = 4.0)
		)
		
		# pretrain transe
		trainer = Trainer(model = model_e, data_loader = dataloader.train_dataloader(),
			epochs = 1, lr = 0.5, opt_method = "sgd", use_gpu = True, device = 'cuda:0')
		trainer.run()
		parameters = transe.get_parameters()
		transe.save_parameters("../../checkpoint/transr_transe.json")
		
		# test the transr
		tester = Tester(model = transr, data_loader = dataloader, use_gpu = True, device = 'cuda:0')
		
		# train transr
		transr.set_parameters(parameters)
		trainer = Trainer(model = model_r, data_loader = dataloader.train_dataloader(),
			epochs = 1000, lr = 1.0, opt_method = "sgd", use_gpu = True, device = 'cuda:0',
			tester = tester, test = True, valid_interval = 100,
			log_interval = 100, save_interval = 100, save_path = '../../checkpoint/transr.pth')
		trainer.run()
		
		# test the model
		transr.load_checkpoint('../../checkpoint/transr.pth')
		tester.set_sampling_mode("link_test")
		tester.run_link_prediction()
	"""

	def __init__(
		self,
		ent_tol: int,
		rel_tol: int,
		dim_e: int = 100,
		dim_r: int = 100,
		p_norm: int = 1,
		norm_flag: bool = True,
		rand_init: bool = False,
		margin: float | None = None):

		"""创建 TransR 对象。

		:param ent_tol: 实体的个数
		:type ent_tol: int
		:param rel_tol: 关系的个数
		:type rel_tol: int
		:param dim_e: 实体嵌入向量的维度
		:type dim_e: int
		:param dim_r: 关系嵌入向量的维度
		:type dim_r: int
		:param p_norm: 评分函数的距离函数, 按照原论文，这里可以取 1 或 2。
		:type p_norm: int
		:param norm_flag: 是否利用 :py:func:`torch.nn.functional.normalize` 对实体和关系嵌入的最后一维执行 L2-norm。
		:type norm_flag: bool
		:param rand_init: 关系矩阵是否采用随机初始化。
		:type rand_init: bool
		:param margin: 当使用 ``RotatE`` :cite:`RotatE` 的损失函数 :py:class:`unike.module.loss.SigmoidLoss`，需要提供此参数，将 ``TransE`` :cite:`TransE` 的正三元组的评分由越小越好转化为越大越好，如果想获得更详细的信息请访问 :ref:`RotatE <rotate>`。
		:type margin: float
		"""

		super(TransR, self).__init__(ent_tol, rel_tol)
		
		#: 实体嵌入向量的维度
		self.dim_e: int = dim_e
		#: 关系嵌入向量的维度
		self.dim_r: int = dim_r
		#: 评分函数的距离函数, 按照原论文，这里可以取 1 或 2。
		self.p_norm: int = p_norm
		#: 是否利用 :py:func:`torch.nn.functional.normalize` 
		#: 对实体和关系嵌入向量的最后一维执行 L2-norm。
		self.norm_flag: bool = norm_flag
		#: 关系矩阵是否采用随机初始化
		self.rand_init: bool = rand_init

		#: 根据实体个数，创建的实体嵌入
		self.ent_embeddings: torch.nn.Embedding = nn.Embedding(self.ent_tol, self.dim_e)
		#: 根据关系个数，创建的关系嵌入
		self.rel_embeddings: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim_r)

		if margin != None:
			#: 当使用 ``RotatE`` :cite:`RotatE` 的损失函数 :py:class:`unike.module.loss.SigmoidLoss`，需要提供此参数，将 ``TransE`` :cite:`TransE` 的正三元组的评分由越小越好转化为越大越好，如果想获得更详细的信息请访问 :ref:`RotatE <rotate>`。
			self.margin: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor([margin]))
			self.margin.requires_grad = False
			self.margin_flag: bool = True
		else:
			self.margin_flag: bool = False

		nn.init.xavier_uniform_(self.ent_embeddings.weight.data)
		nn.init.xavier_uniform_(self.rel_embeddings.weight.data)

		#: 关系矩阵
		self.transfer_matrix: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim_e * self.dim_r)

		if not self.rand_init:
			identity = torch.zeros(self.dim_e, self.dim_r)
			for i in range(min(self.dim_e, self.dim_r)):
				identity[i][i] = 1
			identity = identity.view(self.dim_e * self.dim_r)
			for i in range(self.rel_tol):
				self.transfer_matrix.weight.data[i] = identity
		else:
			nn.init.xavier_uniform_(self.transfer_matrix.weight.data)

	@override
	def forward(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> torch.Tensor:

		"""
		定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
		rel_transfer = self.transfer_matrix(triples[:, 1])
		head_emb = self._transfer(head_emb, rel_transfer)
		tail_emb = self._transfer(tail_emb, rel_transfer)
		score = self._calc(head_emb, relation_emb, tail_emb)
		if self.margin_flag:
			return self.margin - score
		else:
			return score

	def _transfer(
		self,
		e: torch.Tensor,
		r_transfer: torch.Tensor) -> torch.Tensor:

		"""
		将头实体或尾实体的向量投影到特定的关系向量空间。
		
		:param e: 头实体或尾实体向量。
		:type e: torch.Tensor
		:param r_transfer: 特定关系矩阵
		:type r_transfer: torch.Tensor
		:returns: 投影后的实体向量
		:rtype: torch.Tensor
		"""

		r_transfer = r_transfer.view(-1, self.dim_e, self.dim_r)
		r_transfer = r_transfer.unsqueeze(dim=1)
		e = e.unsqueeze(dim=-2)
		e = torch.matmul(e, r_transfer)
		return e.squeeze(dim=-2)

	def _calc(
		self,
		h: torch.Tensor,
		r: torch.Tensor,
		t: torch.Tensor) -> torch.Tensor:

		"""计算 TransR 的评分函数。
		
		:param h: 头实体的向量。
		:type h: torch.Tensor
		:param r: 关系的向量。
		:type r: torch.Tensor
		:param t: 尾实体的向量。
		:type t: torch.Tensor
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		# 对嵌入的最后一维进行归一化
		if self.norm_flag:
			h = F.normalize(h, 2, -1)
			r = F.normalize(r, 2, -1)
			t = F.normalize(t, 2, -1)
		
		score = (h + r) - t
		
		# 利用距离函数计算得分
		score = torch.norm(score, self.p_norm, -1)
		return score

	@override
	def predict(
		self,
		data: dict[str, typing.Union[torch.Tensor,str]],
		mode: str) -> torch.Tensor:
		
		"""TransR 的推理方法。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor,str]]
		:param mode: 'head_predict' 或 'tail_predict'
		:type mode: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		triples = data["positive_sample"]
		head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
		rel_transfer = self.transfer_matrix(triples[:, 1])
		head_emb = self._transfer(head_emb, rel_transfer)
		tail_emb = self._transfer(tail_emb, rel_transfer)
		score = self._calc(head_emb, relation_emb, tail_emb)

		if self.margin_flag:
			score = self.margin - score
			return score
		else:
			return -score

	def regularization(
		self,
		data: dict[str, typing.Union[torch.Tensor, str]]) -> torch.Tensor:

		"""L2 正则化函数（又称权重衰减），在损失函数中用到。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor, str]]
		:returns: 模型参数的正则损失
		:rtype: torch.Tensor
		"""

		pos_sample = data["positive_sample"]
		neg_sample = data["negative_sample"]
		mode = data["mode"]
		pos_head_emb, pos_relation_emb, pos_tail_emb = self.tri2emb(pos_sample)
		pos_rel_transfer = self.transfer_matrix(pos_sample[:, 1])
		if mode == "bern":
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(neg_sample)
		else:
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(pos_sample, neg_sample, mode)
		neg_rel_transfer = self.transfer_matrix(pos_sample[:, 1])

		pos_regul = (torch.mean(pos_head_emb ** 2) + 
					 torch.mean(pos_relation_emb ** 2) + 
					 torch.mean(pos_tail_emb ** 2) +
					 torch.mean(pos_rel_transfer ** 2)) / 4

		neg_regul = (torch.mean(neg_head_emb ** 2) + 
					 torch.mean(neg_relation_emb ** 2) + 
					 torch.mean(neg_tail_emb ** 2) +
					 torch.mean(neg_rel_transfer ** 2)) / 4

		regul = (pos_regul + neg_regul) / 2

		return regul

def get_transr_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`TransR` 的默认超参数优化配置。
	
	``TransR`` :cite:`TransR` 进行超参数优化的时候，需要先训练一个 ``TransE`` :cite:`TransE` 模型（训练 1 epoch）。
	然后 ``TransR`` :cite:`TransR` 的实体和关系的嵌入向量初始化为 TransE 的结果。
	**margin_e** 、 **lr_e** 和 **opt_method_e** 是 ``TransE`` :cite:`TransE` 的训练超参数。
	如果想获得更详细的信息请访问 :ref:`TransR <transr>`。
	
	默认配置为::
	
		parameters_dict = {
			'model': {
				'value': 'TransR'
			},
			'dim': {
				'values': [50, 100]
			},
			'p_norm': {
				'values': [1, 2]
			},
			'norm_flag': {
				'value': True
			},
			'rand_init': {
				'value': False
			},
			'margin_e': {
				'values': [1.0, 3.0, 6.0]
			},
			'lr_e': {
				'distribution': 'uniform',
				'min': 1e-5,
				'max': 1.0
			},
			'opt_method_e': {
				'values': ['adam', 'adagrad', 'sgd']
			},
		}

	:returns: :py:class:`TransR` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'model': {
			'value': 'TransR'
		},
		'dim': {
			'values': [50, 100]
		},
		'p_norm': {
			'values': [1, 2]
		},
		'norm_flag': {
			'value': True
		},
		'rand_init': {
			'value': False
		},
		'margin_e': {
			'values': [1.0, 3.0, 6.0]
		},
		'lr_e': {
			'distribution': 'uniform',
			'min': 1e-5,
			'max': 1.0
		},
		'opt_method_e': {
			'values': ['adam', 'adagrad', 'sgd']
		},
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/module/model/RotatE.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 11, 2023
# 
# 该头文件定义了 RotatE.

"""
RotatE - 将实体表示成复数向量，关系建模为复数向量空间的旋转。
"""

import torch
import typing
import torch.nn as nn
from .Model import Model
from typing_extensions import override

class RotatE(Model):

	"""
	``RotatE`` :cite:`RotatE` 提出于 2019 年，将实体表示成复数向量，关系建模为复数向量空间的旋转。

	评分函数为:

	.. math::
	
		\gamma - \parallel \mathbf{h} \circ \mathbf{r} - \mathbf{t} \parallel_{L_2}
	
	:math:`\circ` 表示哈达玛积（Hadamard product），正三元组的评分函数的值越大越好，如果想获得更详细的信息请访问 :ref:`RotatE <rotate>`。

	例子::

		from unike.data import KGEDataLoader, UniSampler, TradTestSampler
		from unike.module.model import RotatE
		from unike.module.loss import SigmoidLoss
		from unike.module.strategy import NegativeSampling
		from unike.config import Trainer, Tester
		
		# dataloader for training
		dataloader = KGEDataLoader(
			in_path = '../../benchmarks/WN18RR/', 
			batch_size = 2000,
			neg_ent = 64,
			test = True,
			test_batch_size = 10,
			num_workers = 16,
			train_sampler = UniSampler,
			test_sampler = TradTestSampler
		)
		
		# define the model
		rotate = RotatE(
			ent_tol = dataloader.get_ent_tol(),
			rel_tol = dataloader.get_rel_tol(),
			dim = 1024,
			margin = 6.0,
			epsilon = 2.0,
		)
		
		# define the loss function
		model = NegativeSampling(
			model = rotate, 
			loss = SigmoidLoss(adv_temperature = 2), 
			regul_rate = 0.0,
		)
		
		# test the model
		tester = Tester(model = rotate, data_loader = dataloader, use_gpu = True, device = 'cuda:1')
		
		# train the model
		trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(), epochs = 6000,
			lr = 2e-5, opt_method = 'adam', use_gpu = True, device = 'cuda:1',
			tester = tester, test = True, valid_interval = 100,
			log_interval = 100, save_interval = 100,
			save_path = '../../checkpoint/rotate.pth', use_wandb = False)
		trainer.run()
	"""

	def __init__(
		self,
		ent_tol: int,
		rel_tol: int,
		dim: int = 100,
		margin: float = 6.0,
		epsilon: float = 2.0):

		"""创建 RotatE 对象。

		:param ent_tol: 实体的个数
		:type ent_tol: int
		:param rel_tol: 关系的个数
		:type rel_tol: int
		:param dim: 实体和关系嵌入向量的维度
		:type dim: int
		:param margin: 原论文中损失函数的 gamma。
		:type margin: float
		:param epsilon: RotatE 原论文对应的源代码固定为 2.0。
		:type epsilon: float
		"""

		super(RotatE, self).__init__(ent_tol, rel_tol)

		#: RotatE 原论文对应的源代码固定为 2.0。
		self.epsilon: int = epsilon

		#: RotatE 原论文的实现中将实体嵌入向量的维度指定为 ``dim`` 的 2 倍。
		#: 因为实体嵌入向量需要划分为实部和虚部。
		self.dim_e: int = dim * 2
		#: 关系嵌入向量的维度，为 ``dim``。
		self.dim_r: int = dim

		#: 根据实体个数，创建的实体嵌入。
		self.ent_embeddings: torch.nn.Embedding = nn.Embedding(self.ent_tol, self.dim_e)
		#: 根据关系个数，创建的关系嵌入。
		self.rel_embeddings: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim_r)

		self.ent_embedding_range = nn.Parameter(
			torch.Tensor([(margin + self.epsilon) / self.dim_e]), 
			requires_grad=False
		)

		nn.init.uniform_(
			tensor = self.ent_embeddings.weight.data, 
			a=-self.ent_embedding_range.item(), 
			b=self.ent_embedding_range.item()
		)

		self.rel_embedding_range = nn.Parameter(
			torch.Tensor([(margin + self.epsilon) / self.dim_r]), 
			requires_grad=False
		)

		nn.init.uniform_(
			tensor = self.rel_embeddings.weight.data, 
			a=-self.rel_embedding_range.item(), 
			b=self.rel_embedding_range.item()
		)

		#: 原论文中损失函数的 gamma。
		self.margin: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor([margin]))
		self.margin.requires_grad = False

	@override
	def forward(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> torch.Tensor:

		"""
		定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
		score = self.margin - self._calc(head_emb, relation_emb, tail_emb)
		return score

	def _calc(
		self,
		h: torch.Tensor,
		r: torch.Tensor,
		t: torch.Tensor) -> torch.Tensor:

		"""计算 RotatE 的评分函数。

		利用 :py:func:`torch.chunk` 拆分实体嵌入向量获得复数的实部和虚部。
		原论文使用 L1-norm 作为距离函数，而这里使用的 L2-norm 作为距离函数。
		
		:param h: 头实体的向量。
		:type h: torch.Tensor
		:param r: 关系的向量。
		:type r: torch.Tensor
		:param t: 尾实体的向量。
		:type t: torch.Tensor
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		pi = self.pi_const

		re_head, im_head = torch.chunk(h, 2, dim=-1)
		re_tail, im_tail = torch.chunk(t, 2, dim=-1)

		# Make phases of relations uniformly distributed in [-pi, pi]
		phase_relation = r / (self.rel_embedding_range.item() / pi)

		re_relation = torch.cos(phase_relation)
		im_relation = torch.sin(phase_relation)
	
		re_score = re_head * re_relation - im_head * im_relation
		im_score = re_head * im_relation + im_head * re_relation
		re_score = re_score - re_tail
		im_score = im_score - im_tail

		score = torch.stack([re_score, im_score], dim = 0)
		score = score.norm(dim = 0).sum(dim = -1)
		return score

	@override
	def predict(
		self,
		data: dict[str, typing.Union[torch.Tensor,str]],
		mode) -> torch.Tensor:
		
		"""RotatE 的推理方法。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor,str]]
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		triples = data["positive_sample"]
		head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
		score = self.margin - self._calc(head_emb, relation_emb, tail_emb)
		return score

	def regularization(
		self,
		data: dict[str, typing.Union[torch.Tensor, str]]) -> torch.Tensor:

		"""L2 正则化函数（又称权重衰减），在损失函数中用到。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor, str]]
		:returns: 模型参数的正则损失
		:rtype: torch.Tensor
		"""

		pos_sample = data["positive_sample"]
		neg_sample = data["negative_sample"]
		mode = data["mode"]
		pos_head_emb, pos_relation_emb, pos_tail_emb = self.tri2emb(pos_sample)
		if mode == "bern":
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(neg_sample)
		else:
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(pos_sample, neg_sample, mode)

		pos_regul = (torch.mean(pos_head_emb ** 2) + 
					 torch.mean(pos_relation_emb ** 2) + 
					 torch.mean(pos_tail_emb ** 2)) / 3

		neg_regul = (torch.mean(neg_head_emb ** 2) + 
					 torch.mean(neg_relation_emb ** 2) + 
					 torch.mean(neg_tail_emb ** 2)) / 3

		regul = (pos_regul + neg_regul) / 2

		return regul

def get_rotate_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`RotatE` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'model': {
				'value': 'RotatE'
			},
			'dim': {
				'values': [256, 512, 1024]
			},
			'margin': {
				'values': [1.0, 3.0, 6.0]
			},
			'epsilon': {
				'value': 2.0
			}
		}

	:returns: :py:class:`RotatE` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'model': {
			'value': 'RotatE'
		},
		'dim': {
			'values': [256, 512, 1024]
		},
		'margin': {
			'values': [1.0, 3.0, 6.0]
		},
		'epsilon': {
			'value': 2.0
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/module/model/SimplE.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 7, 2023
# 
# 该头文件定义了 SimplE.

"""
SimplE - 简单的双线性模型，能够为头实体和尾实体学习不同的嵌入向量。
"""

import math
import torch
import typing
import numpy as np
import torch.nn as nn
from .Model import Model
from typing_extensions import override

class SimplE(Model):
    
    """
    ``SimplE`` :cite:`SimplE` 提出于 2018 年，简单的双线性模型，能够为头实体和尾实体学习不同的嵌入向量。
    
    评分函数为: 

    .. math::
    
        1/2(<\mathbf{h}_{i}, \mathbf{v}_r, \mathbf{t}_{j}> + <\mathbf{h}_{j}, \mathbf{v}_{r^{-1}}, \mathbf{t}_{i}>)
    
    :math:`< \mathbf{a}, \mathbf{b}, \mathbf{c} >` 为逐元素多线性点积（element-wise multi-linear dot product）。

    正三元组的评分函数的值越大越好，负三元组越小越好，如果想获得更详细的信息请访问 :ref:`SimplE <simple>`。

    例子::

        from unike.config import Trainer, Tester
        from unike.module.model import SimplE
        from unike.module.loss import SoftplusLoss
        from unike.module.strategy import NegativeSampling

        # define the model
        simple = SimplE(
        	ent_tol = train_dataloader.get_ent_tol(),
        	rel_tol = train_dataloader.get_rel_tol(),
        	dim = config.dim
        )

        # define the loss function
        model = NegativeSampling(
        	model = simple, 
        	loss = SoftplusLoss(),
        	batch_size = train_dataloader.get_batch_size(), 
        	regul_rate = config.regul_rate
        )

        # dataloader for test
        test_dataloader = TestDataLoader(in_path = config.in_path)

        # test the model
        tester = Tester(model = simple, data_loader = test_dataloader, use_gpu = config.use_gpu, device = config.device)

        # train the model
        trainer = Trainer(model = model, data_loader = train_dataloader, epochs = config.epochs,
        	lr = config.lr, opt_method = config.opt_method, use_gpu = config.use_gpu, device = config.device,
        	tester = tester, test = config.test, valid_interval = config.valid_interval,
        	log_interval = config.log_interval, save_interval = config.save_interval,
        	save_path = config.save_path, wandb_logger = wandb_logger)
        trainer.run()
    """
    
    def __init__(
        self,
        ent_tol: int,
        rel_tol: int,
        dim: int = 100):

        """创建 SimplE 对象。
        
        :param ent_tol: 实体的个数
        :type ent_tol: int
        :param rel_tol: 关系的个数
        :type rel_tol: int
        :param dim: 实体嵌入向量和关系嵌入向量的维度
        :type dim: int
        """
        
        super(SimplE, self).__init__(ent_tol, rel_tol)
        
        #: 实体嵌入向量和关系嵌入向量的维度
        self.dim: int = dim

        #: 根据实体个数，创建的实体嵌入
        self.ent_embeddings: torch.nn.Embedding = nn.Embedding(self.ent_tol, self.dim * 2)
        #: 根据关系个数，创建的关系嵌入
        self.rel_embeddings: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim * 2)

        nn.init.xavier_uniform_(self.ent_embeddings.weight.data)
        nn.init.xavier_uniform_(self.rel_embeddings.weight.data)

    @override
    def forward(
        self,
        triples: torch.Tensor,
        negs: torch.Tensor = None,
        mode: str = 'single') -> torch.Tensor:
        
        """
        定义每次调用时执行的计算。
        :py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
        
        :param triples: 正确的三元组
        :type triples: torch.Tensor
        :param negs: 负三元组类别
        :type negs: torch.Tensor
        :param mode: 模式
        :type triples: str
        :returns: 三元组的得分
        :rtype: torch.Tensor
		"""

        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        score = self._calc(head_emb, relation_emb, tail_emb)
        return score

    def _calc(
        self,
        h: torch.Tensor,
        r: torch.Tensor,
        t: torch.Tensor) -> torch.Tensor:
        
        """计算 SimplE 的评分函数。
        
        :param h: 头实体的向量。
        :type h: torch.Tensor
        :param r: 关系的向量。
        :type r: torch.Tensor
        :param t: 尾实体的向量。
        :type t: torch.Tensor
        :returns: 三元组的得分
        :rtype: torch.Tensor
        """

        hh_embs, th_embs = torch.chunk(h, 2, dim=-1)
        r_embs, r_inv_embs = torch.chunk(r, 2, dim=-1)
        ht_embs, tt_embs = torch.chunk(t, 2, dim=-1)

        scores1 = torch.sum(hh_embs * r_embs * tt_embs, -1)
        scores2 = torch.sum(ht_embs * r_inv_embs * th_embs, -1)
        
        # Without clipping, we run into NaN problems.
        # 基于论文作者的实现。
        return torch.clamp((scores1 + scores2) / 2, -20, 20)

    @override
    def predict(
        self,
        data: dict[str, typing.Union[torch.Tensor,str]],
        mode) -> torch.Tensor:
        
        """SimplE 的推理方法。
        
        :param data: 数据。
        :type data: dict[str, typing.Union[torch.Tensor,str]]
        :returns: 三元组的得分
        :rtype: torch.Tensor
        """

        triples = data["positive_sample"]
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        score = self._calc(head_emb, relation_emb, tail_emb)
        return score

    def regularization(
        self,
        data: dict[str, typing.Union[torch.Tensor, str]]) -> torch.Tensor:

        """L2 正则化函数（又称权重衰减），在损失函数中用到。
        
        :param data: 数据。
        :type data: dict[str, typing.Union[torch.Tensor, str]]
        :returns: 模型参数的正则损失
        :rtype: torch.Tensor
        """

        pos_sample = data["positive_sample"]
        neg_sample = data["negative_sample"]
        mode = data["mode"]
        pos_head_emb, pos_relation_emb, pos_tail_emb = self.tri2emb(pos_sample)
        if mode == "bern":
            neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(neg_sample)
        else:
            neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(pos_sample, neg_sample, mode)
            
        pos_regul = (torch.mean(pos_head_emb ** 2) + 
                     torch.mean(pos_relation_emb ** 2) + 
                     torch.mean(pos_tail_emb ** 2)) / 3
                     
        neg_regul = (torch.mean(neg_head_emb ** 2) + 
                     torch.mean(neg_relation_emb ** 2) + 
                     torch.mean(neg_tail_emb ** 2)) / 3
                     
        regul = (pos_regul + neg_regul) / 2

        return regul

def get_simple_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`SimplE` 的默认超参数优化配置。
	
	默认配置为::
	
	    parameters_dict = {
	    	'model': {
	    		'value': 'SimplE'
	    	},
	    	'dim': {
	    		'values': [50, 100, 200]
	    	}
	    }

	:returns: :py:class:`SimplE` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'model': {
			'value': 'SimplE'
		},
		'dim': {
			'values': [50, 100, 200]
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/module/__init__.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 28, 2023
# 
# 该头文件定义了 module 接口.

"""模块部分，包含模型和损失函数。"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .BaseModule import BaseModule

__init__ = {
	'BaseModule'
}


# coding:utf-8
#
# unike/utils/Timer.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on July 6, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 24, 2023
#
# 该脚本定义了计时器类.

"""
:py:class:`Timer` - 计时器类。
"""

import time

class Timer:

    """记录多次实验的时间。 
    
    通过调用 :py:meth:`stop` 能够返回距离上一次 :py:meth:`stop` 调用或创建 :py:class:`Timer` 对象时的时间间隔。
    
    :py:meth:`avg` 能够返回多次实验的平均时间；:py:meth:`sum` 能够返回多次实验的总时间。"""

    def __init__(self):

        """创建 Timer 对象。"""

        #: 存放时间间隔的列表
        self.times: list[float] = []
        #: 记录当前时间
        self.current: float = None
        #: 记录上一次的时间
        self.last: float = None

        self.__restart()

    def __restart(self):

        """重启计时器。"""

        self.last = self.current = time.time()

    def stop(self) -> float:

        """停止计时器并将时间记录在列表中。
        
        :returns: 返回最后一次的间隔时间。
        :rtype: float
        """
        
        self.current = time.time()
        self.times.append(self.current - self.last)
        self.last = self.current
        return self.times[-1]

    def avg(self) -> float:

        """返回平均时间。
        
        :returns: 平均时间
        :rtype: float
        """
        
        return sum(self.times) / len(self.times)

    def sum(self) -> float:

        """返回时间总和。
        
        :returns: 时间总和。
        :rtype: float
        """
        
        return sum(self.times)


# coding:utf-8
#
# unike/module/model/CompGCN.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 19, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 26, 2024
#
# 该脚本定义了 CompGCN 类.

"""
CompGCN - 这是一种在图卷积网络中整合多关系信息的新框架，它利用知识图谱嵌入技术中的各种组合操作，将实体和关系共同嵌入到图中。
"""

import dgl
import torch
import typing
from torch import nn
import dgl.function as fn
from .Model import Model
import torch.nn.functional as F
from typing_extensions import override

class CompGCN(Model):

    """
    ``CompGCN`` :cite:`CompGCN` 发表于 ``2020`` 年，这是一种在图卷积网络中整合多关系信息的新框架，它利用知识图谱嵌入技术中的各种组合操作，将实体和关系共同嵌入到图中。

    正三元组的评分函数的值越大越好，如果想获得更详细的信息请访问 :ref:`CompGCN <compgcn>`。

    例子::

        from unike.module.model import CompGCN
        from unike.module.loss import CompGCNLoss
        from unike.module.strategy import CompGCNSampling
        from unike.config import Trainer, GraphTester
        
        # define the model
        compgcn = CompGCN(
        	ent_tol = dataloader.get_ent_tol(),
        	rel_tol = dataloader.get_rel_tol(),
        	dim = 100
        )
        
        # define the loss function
        model = CompGCNSampling(
        	model = compgcn,
        	loss = CompGCNLoss(model = compgcn),
        	ent_tol = dataloader.get_ent_tol()
        )
        
        # test the model
        tester = GraphTester(model = compgcn, data_loader = dataloader, use_gpu = True, device = 'cuda:0', prediction = "tail")
        
        # train the model
        trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(),
        	epochs = 2000, lr = 0.0001, use_gpu = True, device = 'cuda:0',
        	tester = tester, test = True, valid_interval = 50, log_interval = 50,
        	save_interval = 50, save_path = '../../checkpoint/compgcn.pth'
        )
        trainer.run()
    """

    def __init__(
        self,
        ent_tol: int,
        rel_tol: int,
        dim: int,
        opn: str = 'mult',
        fet_drop: float = 0.2,
        hid_drop: float = 0.3,
        margin: float = 40.0,
        decoder_model: str = 'ConvE'):

        """创建 RGCN 对象。
        
        :param ent_tol: 实体的个数
        :type ent_tol: int
        :param rel_tol: 关系的个数
        :type rel_tol: int
        :param dim: 实体和关系嵌入向量的维度
        :type dim: int
        :param opn: 组成运算符：'mult'、'sub'、'corr'
        :type opn: str
        :param fet_drop: 用于 'ConvE' 解码器，用于卷积特征的 dropout
        :type fet_drop: float
        :param hid_drop: 用于 'ConvE' 解码器，用于隐藏层的 dropout
        :type hid_drop: float
        :param margin: 用于 'TransE' 解码器，gamma。
        :type margin: float
        :param decoder_model: 用什么得分函数作为解码器: 'ConvE'、'DistMult'、'TransE'
        :type decoder_model: str
		"""

        super(CompGCN, self).__init__(ent_tol, rel_tol)

        #: 实体和关系嵌入向量的维度
        self.dim: int = dim
        #: 组成运算符：'mult'、'sub'、'corr'
        self.opn: str = opn
        #: 用什么得分函数作为解码器: 'ConvE'、'DistMult'
        self.decoder_model: str = decoder_model

        #------------------------------CompGCN--------------------------------------------------------------------
        #: 根据实体个数，创建的实体嵌入
        self.ent_emb: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor(self.ent_tol, self.dim))
        #: 根据关系个数，创建的关系嵌入
        self.rel_emb: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor(self.rel_tol, self.dim))

        nn.init.xavier_normal_(self.ent_emb, gain=nn.init.calculate_gain('relu'))
        nn.init.xavier_normal_(self.rel_emb, gain=nn.init.calculate_gain('relu'))

        #: CompGCNCov
        self.GraphCov: CompGCNCov = CompGCNCov(self.dim, self.dim * 2, torch.tanh, bias = 'False', drop_rate = 0.1, opn = self.opn)
        #: 用于 :py:attr:`GraphCov` 输出结果
        self.drop: torch.nn.Dropout = nn.Dropout(0.3)
        #: 最后计算得分时的偏置
        self.bias: torch.nn.parameter.Parameter = nn.Parameter(torch.zeros(self.ent_tol))
        #-----------------------------ConvE-----------------------------------------------------------------------
        #: 用于 'ConvE' 解码器，头实体嵌入向量和关系嵌入向量的 BatchNorm
        self.bn0: torch.nn.BatchNorm2d = torch.nn.BatchNorm2d(1)
        #: 用于 'ConvE' 解码器，卷积层
        self.conv1: torch.nn.Conv2d = torch.nn.Conv2d(1, 200, (7, 7), 1, 0, bias=False)
        #: 用于 'ConvE' 解码器，卷积特征的 BatchNorm
        self.bn1: torch.nn.Conv2d = torch.nn.BatchNorm2d(200)
        #: 用于 'ConvE' 解码器，卷积特征的 Dropout
        self.fet_drop: torch.nn.Dropout = torch.nn.Dropout2d(fet_drop)
        flat_sz_h = 4 * self.dim // 20 - 7 + 1
        flat_sz_w = 20 - 7 + 1
        flat_sz = flat_sz_h * flat_sz_w * 200
        #: 用于 'ConvE' 解码器，隐藏层层
        self.fc: torch.nn.Linear = torch.nn.Linear(flat_sz, self.dim*2)
        #: 用于 'ConvE' 解码器，隐藏层的 Dropout
        self.hid_drop: torch.nn.Dropout = torch.nn.Dropout(hid_drop)
        #: 用于 'ConvE' 解码器，隐藏层的 BatchNorm
        self.bn2: torch.nn.BatchNorm1d = torch.nn.BatchNorm1d(self.dim*2)
        #-----------------------------TransE-----------------------------------------------------------------------
        #: 用于 TransE 得分函数
        self.margin: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor([margin]))
        self.margin.requires_grad = False

    @override
    def forward(
        self,
        graph: dgl.DGLGraph,
        relation: torch.Tensor,
        norm: torch.Tensor,
        triples: torch.Tensor) -> torch.Tensor:

        """
		定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
        
        :param graph: 子图
        :type graph: dgl.DGLGraph
        :param relation: 子图的关系
        :type relation: torch.Tensor
        :param norm: 关系的归一化系数
        :type norm: torch.Tensor
        :param triples: 三元组
        :type triples: torch.Tensor
        :returns: 三元组的得分
        :rtype: torch.Tensor
		"""

        head, rela = triples[:,0], triples[:, 1]
        x, r = self.ent_emb, self.rel_emb
        x, r = self.GraphCov(graph, x, r, relation, norm)
        x = self.drop(x)
        head_emb = torch.index_select(x, 0, head)
        rela_emb = torch.index_select(r, 0, rela)

        if self.decoder_model.lower() == 'conve':
           score = self.conve(head_emb, rela_emb, x)
        elif self.decoder_model.lower() == 'distmult':
            score = self.distmult(head_emb, rela_emb, x)
        elif self.decoder_model.lower() == 'transe':
            score = self.transe(head_emb, rela_emb, x)
        else:
            raise ValueError("please choose decoder (TransE/DistMult/ConvE)")

        return score

    def conve(
        self,
        sub_emb: torch.Tensor,
        rel_emb: torch.Tensor,
        all_ent: torch.Tensor) -> torch.Tensor:

        """计算 ConvE 作为解码器时三元组的得分。
        
        :param sub_emb: 头实体的嵌入向量
        :type sub_emb: torch.Tensor
        :param rel_emb: 关系的嵌入向量
        :type rel_emb: torch.Tensor
        :param all_ent: 全部实体的嵌入向量
        :type all_ent: torch.Tensor
        :returns: 三元组的得分
        :rtype: torch.Tensor"""

        stack_input = self.concat(sub_emb, rel_emb)
        x = self.bn0(stack_input)
        x = self.conv1(x)
        x = self.bn1(x)
        x = F.relu(x)
        x = self.fet_drop(x)
        x = x.view(x.shape[0], -1)
        x = self.fc(x)
        x = self.hid_drop(x)
        x = self.bn2(x)
        x = F.relu(x)
        x = torch.mm(x, all_ent.transpose(1, 0))
        x += self.bias.expand_as(x)
        score = torch.sigmoid(x)
        return score

    def concat(
        self,
        ent_embed: torch.Tensor,
        rel_embed: torch.Tensor) -> torch.Tensor:

        """ConvE 作为解码器时，用于拼接头实体嵌入向量和关系嵌入向量。
        
        :param ent_embed: 头实体的嵌入向量
        :type ent_embed: torch.Tensor
        :param rel_embed: 关系的嵌入向量
        :type rel_embed: torch.Tensor
        :returns: ConvE 解码器的输入特征
        :rtype: torch.Tensor"""

        ent_embed = ent_embed.view(-1, 1, self.dim*2)
        rel_embed = rel_embed.view(-1, 1, self.dim*2)
        stack_input = torch.cat([ent_embed, rel_embed], 1)
        stack_input = stack_input.reshape(-1, 1, 4*self.dim//20, 20)
        return stack_input

    def distmult(
        self,
        head_emb: torch.Tensor,
        rela_emb: torch.Tensor,
        all_ent: torch.Tensor) -> torch.Tensor:

        """计算 DistMult 作为解码器时三元组的得分。
        
        :param sub_emb: 头实体的嵌入向量
        :type sub_emb: torch.Tensor
        :param rel_emb: 关系的嵌入向量
        :type rel_emb: torch.Tensor
        :param all_ent: 全部实体的嵌入向量
        :type all_ent: torch.Tensor
        :returns: 三元组的得分
        :rtype: torch.Tensor"""

        obj_emb = head_emb * rela_emb
        x = torch.mm(obj_emb, all_ent.transpose(1, 0))
        x += self.bias.expand_as(x)
        score = torch.sigmoid(x)
        return score

    def transe(
        self,
        head_emb: torch.Tensor,
        rela_emb: torch.Tensor,
        all_ent: torch.Tensor) -> torch.Tensor:

        """计算 TransE 作为解码器时三元组的得分。
        
        :param sub_emb: 头实体的嵌入向量
        :type sub_emb: torch.Tensor
        :param rel_emb: 关系的嵌入向量
        :type rel_emb: torch.Tensor
        :param all_ent: 全部实体的嵌入向量
        :type all_ent: torch.Tensor
        :returns: 三元组的得分
        :rtype: torch.Tensor"""

        obj_emb = head_emb + rela_emb
        x = self.margin - torch.norm(obj_emb.unsqueeze(1) - all_ent, p=1, dim=2)
        score = torch.sigmoid(x)
        return score

    @override
    def predict(
        self,
        data: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]],
        mode: str) -> torch.Tensor:

        """CompGCN 的推理方法。
        
        :param data: 数据。
        :type data: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]
        :param mode: 在 CompGCN 时，无用，只为了保证推理函数形式一致
        :type mode: str
        :returns: 三元组的得分
        :rtype: torch.Tensor
		"""
        
        triples    = data['positive_sample']
        graph      = data['graph']
        relation   = data['rela']
        norm       = data['norm'] 

        head, rela = triples[:,0], triples[:, 1]
        x, r = self.ent_emb, self.rel_emb
        x, r = self.GraphCov(graph, x, r, relation, norm)
        x = self.drop(x)
        head_emb = torch.index_select(x, 0, head)
        rela_emb = torch.index_select(r, 0, rela)

        if self.decoder_model.lower() == 'conve':
           score = self.conve(head_emb, rela_emb, x)
        elif self.decoder_model.lower() == 'distmult':
            score = self.distmult(head_emb, rela_emb, x)
        elif self.decoder_model.lower() == 'transe':
            score = self.transe(head_emb, rela_emb, x)
        else:
            raise ValueError("please choose decoder (TransE/DistMult/ConvE)")

        return score

class CompGCNCov(nn.Module):

    """``CompGCN`` :cite:`CompGCN` 图神经网络模块。"""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        act: typing.Callable[[torch.Tensor], torch.Tensor] = lambda x: x,
        bias: bool = True,
        drop_rate: float = 0.,
        opn: str = 'corr'):

        """创建 CompGCN 对象。
        
        :param in_channels: 输入的特征维度
        :type in_channels: int
        :param out_channels: 输出的特征维度
        :type out_channels: int
        :param act: 激活函数
        :type act: typing.Callable[[torch.Tensor], torch.Tensor]
        :param bias: 是否有偏置
        :type bias: bool
        :param drop_rate: Dropout rate
        :type drop_rate: float
        :param opn: 组成运算符：'mult'、'sub'、'corr'
        :type opn: str
		"""

        super(CompGCNCov, self).__init__()

        #: 输入的特征维度
        self.in_channels: int = in_channels
        #: 输出的特征维度
        self.out_channels: int = out_channels
        self.rel_wt = None
        #: 关系嵌入向量
        self.rel: torch.nn.parameter.Parameter = None
        #: 组成运算符：'mult'、'sub'、'corr'
        self.opn: str = opn
        #：图神经网络的权重矩阵，用于原始关系
        self.in_w: torch.nn.parameter.Parameter = self.get_param([in_channels, out_channels])
        #：图神经网络的权重矩阵，用于相反关系
        self.out_w: torch.nn.parameter.Parameter = self.get_param([in_channels, out_channels])
        #: 用于原始关系和相反关系转换后输出结果的 Dropout
        self.drop: torch.nn.Dropout = nn.Dropout(drop_rate)
        #: 自循环关系嵌入向量的转换矩阵
        self.loop_rel: torch.nn.parameter.Parameter = self.get_param([1, in_channels])
        #：图神经网络的权重矩阵，用于自循环关系
        self.loop_w: torch.nn.parameter.Parameter = self.get_param([in_channels, out_channels])
        #: 偏置
        self.bias: torch.nn.Parameter = nn.Parameter(torch.zeros(out_channels)) if bias else None
        #: BatchNorm
        self.bn: torch.nn.BatchNorm1d = torch.nn.BatchNorm1d(out_channels)
        #: 激活函数
        self.act: typing.Callable[[torch.Tensor], torch.Tensor] = act
        #: 关系嵌入向量的转换矩阵
        self.w_rel: torch.nn.parameter.Parameter = self.get_param([in_channels, out_channels])

    def get_param(
        self,
        shape: list[int]) -> torch.nn.parameter.Parameter:

        """获得权重矩阵。
        
        :param shape: 权重矩阵的 shape
        :type shape: list[int]
        :returns: 权重矩阵
        :rtype: torch.nn.parameter.Parameter
        """

        param = nn.Parameter(torch.Tensor(*shape))
        nn.init.xavier_normal_(param, gain=nn.init.calculate_gain('relu'))
        return param

    def forward(
        self,
        graph: dgl.DGLGraph,
        ent_emb: torch.nn.parameter.Parameter,
        rel_emb: torch.nn.parameter.Parameter,
        edge_type: torch.Tensor,
        edge_norm: torch.Tensor) -> tuple[torch.nn.parameter.Parameter, torch.nn.parameter.Parameter]:
        
        """
        定义每次调用时执行的计算。
        :py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
        
        :param graph: 子图
        :type graph: dgl.DGLGraph
        :param ent_emb: 实体嵌入向量
        :type ent_emb: torch.nn.parameter.Parameter
        :param rel_emb: 关系嵌入向量
        :type rel_emb: torch.nn.parameter.Parameter
        :param edge_type: 关系 ID
        :type edge_type: torch.Tensor
        :param norm: 关系的归一化系数
        :type norm: torch.Tensor
        :returns: 更新后的实体嵌入和关系嵌入
        :rtype: tuple[torch.nn.parameter.Parameter, torch.nn.parameter.Parameter]
        """

        graph = graph.local_var()
        graph.ndata['h'] = ent_emb
        graph.edata['type'] = edge_type
        graph.edata['norm'] = edge_norm
        if self.rel_wt is None:
            self.rel = rel_emb
        else:
            self.rel = torch.mm(self.rel_wt, rel_emb)
        graph.update_all(self.message_func, fn.sum(msg='msg', out='h'), self.reduce_func)
        ent_emb = graph.ndata.pop('h') + torch.mm(self.comp(ent_emb, self.loop_rel), self.loop_w) / 3
        if self.bias is not None:
            ent_emb = ent_emb + self.bias
        ent_emb = self.bn(ent_emb)

        return self.act(ent_emb), torch.matmul(self.rel, self.w_rel)

    def message_func(self, edges: dgl.udf.EdgeBatch):

        """
        消息函数。
        """

        edge_type = edges.data['type']
        edge_num = edge_type.shape[0]
        edge_data = self.comp(edges.src['h'], self.rel[edge_type])
        msg = torch.cat([torch.matmul(edge_data[:edge_num // 2, :], self.in_w),
                         torch.matmul(edge_data[edge_num // 2:, :], self.out_w)])
        msg = msg * edges.data['norm'].reshape(-1, 1)
        return {'msg': msg}

    def comp(
        self,
        h: torch.Tensor,
        r: torch.Tensor) -> torch.Tensor:

        """组成运算：'mult'、'sub'、'corr'
        
        :param h: 头实体嵌入向量
        :type h: torch.Tensor
        :param r: 关系嵌入向量
        :type r: torch.Tensor
        :returns: 组合后的边数据
        :rtype: torch.Tensor
        """

        def com_mult(a, b):

            """复数乘法"""

            r1, i1 = a.real, a.imag
            r2, i2 = b.real, b.imag
            real = r1 * r2 - i1 * i2
            imag = r1 * i2 + i1 * r2
            return torch.complex(real, imag)

        def conj(a):

            """共轭复数"""

            a.imag = -a.imag
            return a

        def ccorr(a, b):

            """corr 运算"""

            return torch.fft.irfft(com_mult(conj(torch.fft.rfft(a)), torch.fft.rfft(b)), a.shape[-1])

        if self.opn == 'mult':
            return h * r
        elif self.opn == 'sub':
            return h - r
        elif self.opn == 'corr':
            return ccorr(h, r.expand_as(h))
        else:
            raise KeyError(f'composition operator {self.opn} not recognized.')

    def reduce_func(self, nodes: dgl.udf.NodeBatch):

        """聚合函数"""
        
        return {'h': self.drop(nodes.data['h']) / 3}

def get_compgcn_hpo_config() -> dict[str, dict[str, typing.Any]]:
    
    """返回 :py:class:`CompGCN` 的默认超参数优化配置。
    
    默认配置为::
	
        parameters_dict = {
            'model': {
                'value': 'CompGCN'
            },
            'dim': {
                'values': [100, 150, 200]
            },
            'opn': {
                'value': 'mult'
            },
            'fet_drop': {
                'value': 0.2
            },
            'hid_drop': {
                'value': 0.3
            },
            'margin': {
                'value': 40.0
            },
            'decoder_model': {
                'value': 'ConvE'
            }
        }
        
    :returns: :py:class:`CompGCN` 的默认超参数优化配置
    :rtype: dict[str, dict[str, typing.Any]]
    """
    
    parameters_dict = {
        'model': {
            'value': 'CompGCN'
        },
        'dim': {
            'values': [100, 150, 200]
        },
        'opn': {
            'value': 'mult'
        },
        'fet_drop': {
            'value': 0.2
        },
        'hid_drop': {
            'value': 0.3
        },
        'margin': {
            'value': 40.0
        },
        'decoder_model': {
            'value': 'ConvE'
        }
    }
    
    return parameters_dict


# coding:utf-8
#
# unike/config/Tester.py
#
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 6, 2024
#
# 该脚本定义了验证模型类.

"""
Tester - 验证模型类，内部使用 ``tqmn`` 实现进度条。
"""

import dgl
import torch
import typing
import collections
import numpy as np
from tqdm import tqdm
from ..module.model import Model
from ..data import KGEDataLoader
from loguru import logger

class Tester(object):

    """
    主要用于 KGE 模型的评估。
    
    例子::
    
        from unike.data import KGEDataLoader, BernSampler, TradTestSampler
        from unike.module.model import TransE
        from unike.module.loss import MarginLoss
        from unike.module.strategy import NegativeSampling
        from unike.config import Trainer, Tester

        # dataloader for training
        dataloader = KGEDataLoader(
        	in_path = "../../benchmarks/FB15K/", 
        	batch_size = 8192,
        	neg_ent = 25,
        	test = True,
        	test_batch_size = 256,
        	num_workers = 16,
        	train_sampler = BernSampler,
        	test_sampler = TradTestSampler
        )

        # define the model
        transe = TransE(
        	ent_tol = dataloader.train_sampler.ent_tol,
        	rel_tol = dataloader.train_sampler.rel_tol,
        	dim = 50, 
        	p_norm = 1, 
        	norm_flag = True)

        # define the loss function
        model = NegativeSampling(
        	model = transe, 
        	loss = MarginLoss(margin = 1.0),
        	regul_rate = 0.01
        )

        # test the model
        tester = Tester(model = transe, data_loader = dataloader, use_gpu = True, device = 'cuda:1')

        # train the model
        trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(),
        	epochs = 1000, lr = 0.01, use_gpu = True, device = 'cuda:1',
        	tester = tester, test = True, valid_interval = 100,
        	log_interval = 100, save_interval = 100,
        	save_path = '../../checkpoint/transe.pth', delta = 0.01)
        trainer.run()
    """

    #: 准备报告的指标 Hit@N 的列表，默认为 [1, 3, 10], 表示报告 Hits@1, Hits@3, Hits@10
    hits: list[int] = [1, 3, 10]

    def __init__(
        self,
        model: Model = None,
        data_loader: KGEDataLoader = None,
        sampling_mode: str = 'link_test',
        prediction: str = "all",
        use_tqdm: bool = True,
        use_gpu: bool = True,
        device: str = "cuda:0",
        only_test: bool = False):

        """创建 Tester 对象。
        
        :param model: KGE 模型
        :type model: unike.module.model.Model
        :param data_loader: py:class:`unike.data.KGEDataLoader`
        :type data_loader: unike.data.KGEDataLoader
        :param sampling_mode: 评估验证集还是测试集：**'link_test'** or **'link_valid'**
        :type sampling_mode: str
        :param prediction: 链接预测模式: **'all'**、**'head'**、**'tail'**
        :type prediction: str
        :param use_tqdm: 是否启用进度条
        :type use_tqdm: bool
        :param use_gpu: 是否使用 gpu
        :type use_gpu: bool
        :param device: 使用哪个 gpu
        :type device: str
        :param only_test: 是否是评估已经训练好的模型
        :type only_test: bool
        """

        #: KGE 模型，即 :py:class:`unike.module.model.Model`
        self.model: Model = model
        #: :py:class:`unike.data.KGEDataLoader`
        self.data_loader: KGEDataLoader = data_loader
        #: :py:class:`unike.data.TestDataLoader` 负采样的方式：**'link_test'** or **'link_valid'**
        self.sampling_mode: str = sampling_mode
        #: 链接预测模式: **'all'**、**'head'**、**'tail'**
        self.prediction: str = prediction
        #: 是否启用进度条
        self.use_tqdm: bool = use_tqdm
        #: 是否使用 gpu
        self.use_gpu: bool = use_gpu
        #: gpu，利用 ``device`` 构造的 :py:class:`torch.device` 对象
        self.device: torch.device = torch.device(device)
        #: 验证数据加载器。
        self.val_dataloader: torch.utils.data.DataLoader = self.data_loader.val_dataloader()
        #: 测试数据加载器。
        self.test_dataloader: torch.utils.data.DataLoader = self.data_loader.test_dataloader()
        
        if self.use_gpu and only_test:
            self.model.cuda(device = self.device)

    def set_hits(
        self,
        new_hits: list[int] = [1, 3, 10]):

        """定义 Hits 指标。
        
        :param new_hits: 准备报告的指标 Hit@N 的列表，默认为 [1, 3, 10], 表示报告 Hits@1, Hits@3, Hits@10
        :type new_hits: list[int]
        """

        tmp = self.hits
        self.hits = new_hits

        logger.info(f"Hits@N 指标由 {tmp} 变为 {self.hits}")

    def to_var(
        self,
        x: torch.Tensor) -> torch.Tensor:

        """根据 :py:attr:`use_gpu` 返回 ``x`` 的张量
        
        :param x: 数据
        :type x: torch.Tensor
        :returns: 张量
        :rtype: torch.Tensor
        """

        if self.use_gpu:
            return x.to(self.device)
        else:
            return x

    def run_link_prediction(self) -> dict[str, float]:
        
        """进行链接预测。

        :returns: 经典指标分别为 MR，MRR，Hits@1，Hits@3，Hits@10
        :rtype: dict[str, float]
        """

        if self.sampling_mode == "link_valid":
            training_range = tqdm(self.val_dataloader) if self.use_tqdm else self.val_dataloader
        elif self.sampling_mode == "link_test":
            training_range = tqdm(self.test_dataloader) if self.use_tqdm else self.test_dataloader
        self.model.eval()
        results = collections.defaultdict(float)
        results_type = collections.defaultdict(float)
        with torch.no_grad():
            for data in training_range:
                data = {key : self.to_var(value) for key, value in data.items()}
                if "head_label_type" in data.keys():
                    ranks, ranks_type = link_predict(data, self.model, prediction=self.prediction)
                    results_type["count_type"] += torch.numel(ranks_type)
                    results_type["mr_type"] += torch.sum(ranks_type).item()
                    results_type["mrr_type"] += torch.sum(1.0 / ranks_type).item()
                    for k in self.hits:
                        results_type['hits@{}_type'.format(k)] += torch.numel(ranks_type[ranks_type <= k])
                else:
                    ranks = link_predict(data, self.model, prediction=self.prediction)
                results["count"] += torch.numel(ranks)
                results["mr"] += torch.sum(ranks).item()
                results["mrr"] += torch.sum(1.0 / ranks).item()
                for k in self.hits:
                    results['hits@{}'.format(k)] += torch.numel(ranks[ranks <= k])

        count = results["count"]
        results = {key : np.around(value / count, decimals=3).item() for key, value in results.items() if key != "count"}
        if "count_type" in results_type.keys():
            count_type = results_type["count_type"]
            results_type = {key : np.around(value / count_type, decimals=3).item() for key, value in results_type.items() if key != "count_type"}
            for key, value in results_type.items():
                results[key] = value
        return results
    
    def set_sampling_mode(self, sampling_mode: str):
        
        """设置 :py:attr:`sampling_mode`
        
        :param sampling_mode: 数据采样模式，**'link_test'** 和 **'link_valid'** 分别表示为链接预测进行测试集和验证集的负采样
        :type sampling_mode: str
        """
        
        self.sampling_mode = sampling_mode

def link_predict(
    batch: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]],
    model: Model,
    prediction: str = "all") -> tuple[torch.Tensor, ...]:

    """
    进行链接预测。
    
    :param batch: 测试数据
    :type batch: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]
    :param model: KGE 模型
    :type model: unike.module.model.Model
    :param prediction: **'all'**, **'head'**, **'tail'**
    :type prediction: str
    :returns: 正确三元组的排名
    :rtype: tuple[torch.Tensor, ...]
    """
    
    if prediction == "all":
        tail_ranks = tail_predict(batch, model)
        head_ranks = head_predict(batch, model)
        if "head_label_type" in batch.keys():
            return torch.cat([tail_ranks[0], head_ranks[0]]).float(), torch.cat([tail_ranks[1], head_ranks[1]]).float()
        else:
            return torch.cat([tail_ranks, head_ranks]).float()
    elif prediction == "head":
        if "head_label_type" in batch.keys():
            ranks, ranks_type = head_predict(batch, model)
            return ranks.float(), ranks_type.float()
        else:
            ranks = head_predict(batch, model)
            return ranks.float()
    elif prediction == "tail":
        if "tail_label_type" in batch.keys():
            ranks, ranks_type = tail_predict(batch, model)
            return ranks.float(), ranks_type.float()
        else:
            ranks = tail_predict(batch, model)
            return ranks.float()

def head_predict(
    batch: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]],
    model: Model) -> tuple[torch.Tensor, ...]:

    """
    进行头实体的链接预测。
    
    :param batch: 测试数据
    :type batch: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]
    :param model: KGE 模型
    :type model: unike.module.model.Model
    :returns: 正确三元组的排名
    :rtype: tuple[torch.Tensor, ...]
    """
    
    pos_triple = batch["positive_sample"]
    idx = pos_triple[:, 0]
    label = batch["head_label"]
    pred_score = model.predict(batch, "head_predict")
    if "head_label_type" in batch.keys():
        label_type = batch["head_label_type"]
        return calc_ranks(idx, label, pred_score), calc_ranks(idx, label_type, pred_score)
    return calc_ranks(idx, label, pred_score)

def tail_predict(
    batch: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]],
    model: Model) -> tuple[torch.Tensor, ...]:

    """
    进行尾实体的链接预测。
    
    :param batch: 测试数据
    :type batch: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]
    :param model: KGE 模型
    :type model: unike.module.model.Model
    :returns: 正确三元组的排名
    :rtype: tuple[torch.Tensor, ...]
    """

    pos_triple = batch["positive_sample"]
    idx = pos_triple[:, 2]
    label = batch["tail_label"]
    pred_score = model.predict(batch, "tail_predict")
    if "tail_label_type" in batch.keys():
        label_type = batch["tail_label_type"]
        return calc_ranks(idx, label, pred_score), calc_ranks(idx, label_type, pred_score)
    return calc_ranks(idx, label, pred_score)

def calc_ranks(
    idx: torch.Tensor,
    label: torch.Tensor,
    pred_score: torch.Tensor) -> torch.Tensor:

    """
    计算三元组的排名。
    
    :param idx: 需要链接预测的实体 ID
    :type idx: torch.Tensor
    :param label: 标签
    :type label: torch.Tensor
    :param pred_score: 三元组的评分
    :type pred_score: torch.Tensor
    :returns: 正确三元组的排名
    :rtype: torch.Tensor
    """

    b_range = torch.arange(pred_score.size()[0])
    target_pred = pred_score[b_range, idx]
    pred_score = torch.where(label.bool(), -torch.ones_like(pred_score) * 10000000, pred_score)
    pred_score[b_range, idx] = target_pred

    ranks = (
        1
        + torch.argsort(
            torch.argsort(pred_score, dim=1, descending=True), dim=1, descending=False
        )[b_range, idx]
    )
    return ranks

def get_tester_hpo_config() -> dict[str, dict[str, typing.Any]]:
    
    """返回 :py:class:`Tester` 的默认超参数优化配置。
    
    默认配置为::
    
        parameters_dict = {
            'tester': {
                'value': 'Tester'
            },
            'prediction': {
                'value': 'all'
            },
            'use_tqdm': {
                'value': False
            },
            'use_gpu_tester': {
                'value': True
            },
            'device_tester': {
                'value': 'cuda:0'
            },
        }

    :returns: :py:class:`Tester` 的默认超参数优化配置
    :rtype: dict[str, dict[str, typing.Any]]  
    """
    
    parameters_dict = {
        'tester': {
            'value': 'Tester'
        },
        'prediction': {
            'value': 'all'
        },
        'use_tqdm': {
            'value': False
        },
        'use_gpu_tester': {
            'value': True
        },
        'device_tester': {
            'value': 'cuda:0'
        },
    }
    
    return parameters_dict


# coding:utf-8
#
# unike/module/model/TransD.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 25, 2023
# 
# 该头文件定义了 TransD.

"""
TransD - 自动生成映射矩阵，简单而且高效，是对 TransR 的改进。
"""

import torch
import typing
import torch.nn as nn
import torch.nn.functional as F
from .Model import Model
from typing_extensions import override

class TransD(Model):

	"""
	``TransD`` :cite:`TransD` 提出于 2015 年，自动生成映射矩阵，简单而且高效，是对 TransR 的改进。

	评分函数为:

	.. math::
	
		\parallel (\mathbf{r}_p \mathbf{h}_p^T + \mathbf{I})\mathbf{h} + \mathbf{r} - (\mathbf{r}_p \mathbf{t}_p^T + \mathbf{I})\mathbf{t} \parallel_{L_1/L_2}

	正三元组的评分函数的值越小越好，如果想获得更详细的信息请访问 :ref:`TransD <transd>`。

	例子::

		from unike.utils import WandbLogger
		from unike.data import KGEDataLoader, BernSampler, TradTestSampler
		from unike.module.model import TransD
		from unike.module.loss import MarginLoss
		from unike.module.strategy import NegativeSampling
		from unike.config import Trainer, Tester
		
		wandb_logger = WandbLogger(
			project="unike",
			name="TransD-FB15K237",
			config=dict(
				in_path = '../../benchmarks/FB15K237/',
				batch_size = 2048,
				neg_ent = 25,
				test = True,
				test_batch_size = 10,
				num_workers = 16,
				dim_e = 200,
				dim_r = 200,
				p_norm = 1,
				norm_flag = True,
				margin = 4.0,
				use_gpu = True,
				device = 'cuda:1',
				epochs = 1000,
				lr = 1.0,
				opt_method = 'sgd',
				valid_interval = 100,
				log_interval = 100,
				save_interval = 100,
				save_path = '../../checkpoint/transd.pth'
			)
		)
		
		config = wandb_logger.config
		
		# dataloader for training
		dataloader = KGEDataLoader(
			in_path = config.in_path, 
			batch_size = config.batch_size,
			neg_ent = config.neg_ent,
			test = config.test,
			test_batch_size = config.test_batch_size,
			num_workers = config.num_workers,
			train_sampler = BernSampler,
			test_sampler = TradTestSampler
		)
		
		# define the model
		transd = TransD(
			ent_tol = dataloader.get_ent_tol(),
			rel_tol = dataloader.get_rel_tol(),
			dim_e = config.dim_e, 
			dim_r = config.dim_r, 
			p_norm = config.p_norm, 
			norm_flag = config.norm_flag)
		
		# define the loss function
		model = NegativeSampling(
			model = transd, 
			loss = MarginLoss(margin = config.margin)
		)
		
		# test the model
		tester = Tester(model = transd, data_loader = dataloader, use_gpu = config.use_gpu, device = config.device)
		
		# train the model
		trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(), epochs = config.epochs,
			lr = config.lr, opt_method = config.opt_method, use_gpu = config.use_gpu, device = config.device,
			tester = tester, test = config.test, valid_interval = config.valid_interval,
			log_interval = config.log_interval, save_interval = config.save_interval,
			save_path = config.save_path, wandb_logger = wandb_logger)
		trainer.run()
		
		# close your wandb run
		wandb_logger.finish()
	"""

	def __init__(
		self,
		ent_tol: int,
		rel_tol: int,
		dim_e: int = 100,
		dim_r: int = 100,
		p_norm: int = 1,
		norm_flag: bool = True,
		margin: float | None = None):
		
		"""创建 TransD 对象。

		:param ent_tol: 实体的个数
		:type ent_tol: int
		:param rel_tol: 关系的个数
		:type rel_tol: int
		:param dim_e: 实体嵌入和实体投影向量的维度
		:type dim_e: int
		:param dim_r: 关系嵌入和关系投影向量的维度
		:type dim_r: int
		:param p_norm: 评分函数的距离函数, 按照原论文，这里可以取 1 或 2。
		:type p_norm: int
		:param norm_flag: 是否利用 :py:func:`torch.nn.functional.normalize` 对实体和关系嵌入的最后一维执行 L2-norm。
		:type norm_flag: bool
		:param margin: 当使用 ``RotatE`` :cite:`RotatE` 的损失函数 :py:class:`unike.module.loss.SigmoidLoss`，需要提供此参数，将 ``TransE`` :cite:`TransE` 的正三元组的评分由越小越好转化为越大越好，如果想获得更详细的信息请访问 :ref:`RotatE <rotate>`。
		:type margin: float
		"""

		super(TransD, self).__init__(ent_tol, rel_tol)
		
		#: 实体嵌入和实体投影向量的维度
		self.dim_e: int = dim_e
		#: 关系嵌入和关系投影向量的维度
		self.dim_r: int = dim_r
		#: 评分函数的距离函数, 按照原论文，这里可以取 1 或 2。
		self.p_norm: int = p_norm
		#: 是否利用 :py:func:`torch.nn.functional.normalize` 
		#: 对实体和关系嵌入向量的最后一维执行 L2-norm。
		self.norm_flag: bool = norm_flag

		#: 根据实体个数，创建的实体嵌入
		self.ent_embeddings: torch.nn.Embedding = nn.Embedding(self.ent_tol, self.dim_e)
		#: 根据关系个数，创建的关系嵌入
		self.rel_embeddings: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim_r)
		#: 根据实体个数，创建的实体投影向量
		self.ent_transfer: torch.nn.Embedding = nn.Embedding(self.ent_tol, self.dim_e)
		#: 根据关系个数，创建的关系投影向量
		self.rel_transfer: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim_r)

		if margin != None:
			#: 当使用 ``RotatE`` :cite:`RotatE` 的损失函数 :py:class:`unike.module.loss.SigmoidLoss`，需要提供此参数，将 ``TransE`` :cite:`TransE` 的正三元组的评分由越小越好转化为越大越好，如果想获得更详细的信息请访问 :ref:`RotatE <rotate>`。
			self.margin: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor([margin]))
			self.margin.requires_grad = False
			self.margin_flag: bool = True
		else:
			self.margin_flag: bool = False

		nn.init.xavier_uniform_(self.ent_embeddings.weight.data)
		nn.init.xavier_uniform_(self.rel_embeddings.weight.data)
		nn.init.xavier_uniform_(self.ent_transfer.weight.data)
		nn.init.xavier_uniform_(self.rel_transfer.weight.data)

	@override
	def forward(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> torch.Tensor:

		"""
		定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		h, r, t = self.tri2emb(triples, negs, mode)
		h_transfer, r_transfer, t_transfer = self.tri2transfer(triples, negs, mode)
		h = self._transfer(h, h_transfer, r_transfer)
		t = self._transfer(t, t_transfer, r_transfer)
		score = self._calc(h, r, t)
		if self.margin_flag:
			return self.margin - score
		else:
			return score

	def tri2transfer(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

		"""
		返回三元组对应的嵌入向量。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 头实体、关系和尾实体的嵌入向量
		:rtype: tuple[torch.Tensor, torch.Tensor, torch.Tensor]
		"""
		
		if mode == "single":
			head_emb = self.ent_transfer(triples[:, 0]).unsqueeze(1)
			relation_emb = self.rel_transfer(triples[:, 1]).unsqueeze(1)
			tail_emb = self.ent_transfer(triples[:, 2]).unsqueeze(1)
			
		elif mode == "head-batch" or mode == "head_predict":
			if negs is None:
				head_emb = self.ent_transfer.weight.data.unsqueeze(0)
			else:
				head_emb = self.ent_transfer(negs)
				
			relation_emb = self.rel_transfer(triples[:, 1]).unsqueeze(1)
			tail_emb = self.ent_transfer(triples[:, 2]).unsqueeze(1)
			
		elif mode == "tail-batch" or mode == "tail_predict": 
			head_emb = self.ent_transfer(triples[:, 0]).unsqueeze(1)
			relation_emb = self.rel_transfer(triples[:, 1]).unsqueeze(1)
			
			if negs is None:
				tail_emb = self.ent_transfer.weight.data.unsqueeze(0)
			else:
				tail_emb = self.ent_transfer(negs)
		
		return head_emb, relation_emb, tail_emb

	def _transfer(
		self,
		e: torch.Tensor,
		e_transfer: torch.Tensor,
		r_transfer: torch.Tensor) -> torch.Tensor:

		"""
		将头实体或尾实体的向量映射到关系向量空间。
		
		:param e: 头实体或尾实体向量。
		:type e: torch.Tensor
		:param e_transfer: 头实体或尾实体的投影向量
		:type e_transfer: torch.Tensor
		:param r_transfer: 关系的投影向量
		:type r_transfer: torch.Tensor
		:returns: 投影后的实体向量
		:rtype: torch.Tensor
		"""
	
		return F.normalize(
			self._resize(e, len(e.size())-1, r_transfer.size()[-1]) + torch.sum(e * e_transfer, -1, True) * r_transfer,
			p = 2, 
			dim = -1
		)

	def _resize(
		self,
		tensor: torch.Tensor,
		axis: int,
		size: int) -> torch.Tensor:

		"""计算实体向量与单位矩阵的乘法，并返回结果向量。

		源代码使用 :py:func:`torch.narrow` 进行缩小向量，
		:py:func:`torch.nn.functional.pad` 进行填充向量。
		
		:param tensor: 实体向量。
		:type tensor: torch.Tensor
		:param axis: 在哪个轴上进行乘法运算。
		:type axis: int
		:param size: 运算结果在 ``axis`` 上的维度大小，一般为关系向量的维度。
		:type size: int
		:returns: 乘法结果的向量
		:rtype: torch.Tensor
		"""

		shape = tensor.size()
		osize = shape[axis]
		if osize == size:
			return tensor
		if (osize > size):
			return torch.narrow(tensor, axis, 0, size)
		paddings = []
		for i in range(len(shape)):
			if i == axis:
				paddings = [0, size - osize] + paddings
			else:
				paddings = [0, 0] + paddings
		return F.pad(tensor, paddings, mode = "constant", value = 0)

	def _calc(
		self,
		h: torch.Tensor,
		r: torch.Tensor,
		t: torch.Tensor) -> torch.Tensor:

		"""计算 TransD 的评分函数。
		
		:param h: 头实体的向量。
		:type h: torch.Tensor
		:param r: 关系的向量。
		:type r: torch.Tensor
		:param t: 尾实体的向量。
		:type t: torch.Tensor
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		# 对嵌入的最后一维进行归一化
		if self.norm_flag:
			h = F.normalize(h, 2, -1)
			r = F.normalize(r, 2, -1)
			t = F.normalize(t, 2, -1)
		
		score = (h + r) - t
		
		# 利用距离函数计算得分
		score = torch.norm(score, self.p_norm, -1)
		return score

	@override
	def predict(
		self,
		data: dict[str, typing.Union[torch.Tensor,str]],
		mode: str) -> torch.Tensor:
		
		"""TransH 的推理方法。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor,str]]
		:param mode: 'head_predict' 或 'tail_predict'
		:type mode: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		triples = data["positive_sample"]
		h, r, t = self.tri2emb(triples, mode=mode)
		h_transfer, r_transfer, t_transfer = self.tri2transfer(triples, mode=mode)
		h = self._transfer(h, h_transfer, r_transfer)
		t = self._transfer(t, t_transfer, r_transfer)
		score = self._calc(h, r, t)

		if self.margin_flag:
			score = self.margin - score
			return score
		else:
			return -score

	def regularization(
		self,
		data: dict[str, typing.Union[torch.Tensor, str]]) -> torch.Tensor:
		
		"""L2 正则化函数（又称权重衰减），在损失函数中用到。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor, str]]
		:returns: 模型参数的正则损失
		:rtype: torch.Tensor
		"""

		pos_sample = data["positive_sample"]
		neg_sample = data["negative_sample"]
		mode = data["mode"]
		pos_h, pos_r, pos_t = self.tri2emb(pos_sample)
		pos_h_transfer, pos_r_transfer, pos_t_transfer = self.tri2transfer(pos_sample)
		if mode == "bern":
			neg_h, neg_r, neg_t = self.tri2emb(neg_sample)
			neg_h_transfer, neg_r_transfer, neg_t_transfer = self.tri2transfer(neg_sample)
		else:
			neg_h, neg_r, neg_t = self.tri2emb(pos_sample, neg_sample, mode)
			neg_h_transfer, neg_r_transfer, neg_t_transfer = self.tri2transfer(pos_sample, neg_sample, mode)

		pos_regul = (torch.mean(pos_h ** 2) + 
					 torch.mean(pos_r ** 2) + 
					 torch.mean(pos_t ** 2) +
					 torch.mean(pos_h_transfer ** 2) + 
					 torch.mean(pos_r_transfer ** 2) +
					 torch.mean(pos_t_transfer ** 2)) / 6

		neg_regul = (torch.mean(neg_h ** 2) + 
					 torch.mean(neg_r ** 2) + 
					 torch.mean(neg_t ** 2) +
					 torch.mean(neg_h_transfer ** 2) + 
					 torch.mean(neg_r_transfer ** 2) +
					 torch.mean(neg_t_transfer ** 2)) / 6

		regul = (pos_regul + neg_regul) / 2

		return regul

def get_transd_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`TransD` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'model': {
				'value': 'TransD'
			},
			'dim_e': {
				'values': [50, 100, 200]
			},
			'dim_r': {
				'values': [50, 100, 200]
			},
			'p_norm': {
				'values': [1, 2]
			},
			'norm_flag': {
				'value': True
			}
		}

	:returns: :py:class:`TransD` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'model': {
			'value': 'TransD'
		},
		'dim_e': {
			'values': [50, 100, 200]
		},
		'dim_r': {
			'values': [50, 100, 200]
		},
		'p_norm': {
			'values': [1, 2]
		},
		'norm_flag': {
			'value': True
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/module/loss/SoftplusLoss.py
#
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 6, 2023
#
# 该脚本定义了 regularized logistic loss 损失函数.

"""
SoftplusLoss - 损失函数类，DistMult 经常应用这种损失函数完成模型学习。
"""

import torch
import numpy as np
import torch.nn as nn
from typing import Any
import torch.nn.functional as F
from .Loss import Loss

class SoftplusLoss(Loss):

	"""
	``ComplEx`` :cite:`ComplEx` 原论文中应用这种损失函数完成模型训练。

	.. Note:: :py:meth:`forward` 中的正样本评分函数的得分应大于负样本评分函数的得分。
	
	例子::

		from unike.module.loss import SoftplusLoss
		from unike.module.strategy import NegativeSampling

		# define the loss function
		model = NegativeSampling(
			model = distmult, 
			loss = SoftplusLoss(),
			batch_size = train_dataloader.get_batch_size(), 
			regul_rate = 1.0
		)
	"""

	def __init__(
		self,
		adv_temperature: float | None = None):

		"""创建 SoftplusLoss 对象。

		:param adv_temperature: RotatE 提出的自我对抗负采样中的温度。
		:type adv_temperature: float
		"""

		super(SoftplusLoss, self).__init__()
		#: 用于代替逻辑函数，类型为 :py:class:`torch.nn.Softplus`，它是 ReLU 函数的平滑近似。
		self.criterion: torch.nn.Softplus = nn.Softplus()
		if adv_temperature != None:
			#: RotatE 提出的自我对抗负采样中的温度。
			self.adv_temperature: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor([adv_temperature]))
			self.adv_temperature.requires_grad = False
			#: 是否启用 RotatE 提出的自我对抗负采样。
			self.adv_flag: bool = True
		else:
			self.adv_flag: bool = False
	
	def get_weights(
		self,
		n_score: torch.Tensor) -> torch.Tensor:

		"""计算 RotatE 提出的自我对抗负采样中的负样本的分布概率。
		
		:param n_score: 负样本评分函数的得分。
		:type n_score: torch.Tensor
		:returns: 自我对抗负采样中的负样本的分布概率
		:rtype: torch.Tensor
		"""

		return F.softmax(n_score * self.adv_temperature, dim = -1).detach()

	def forward(
		self,
		p_score: torch.Tensor,
		n_score: torch.Tensor) -> torch.Tensor:

		"""计算 SoftplusLoss 损失函数。定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param p_score: 正样本评分函数的得分。
		:type p_score: torch.Tensor
		:param n_score: 负样本评分函数的得分。
		:type n_score: torch.Tensor
		:returns: 损失值
		:rtype: torch.Tensor
		"""

		if self.adv_flag:
			return (self.criterion(-p_score).mean() + (self.get_weights(n_score) * self.criterion(n_score)).sum(dim = -1).mean()) / 2
		else:
			return (self.criterion(-p_score).mean() + self.criterion(n_score).mean()) / 2
			
	def predict(
		self,
		p_score: torch.Tensor,
		n_score: torch.Tensor) -> np.ndarray:

		"""SoftplusLoss 的推理方法。
		
		:param p_score: 正样本评分函数的得分。
		:type p_score: torch.Tensor
		:param n_score: 负样本评分函数的得分。
		:type n_score: torch.Tensor
		:returns: 损失值
		:rtype: numpy.ndarray
		"""

		score = self.forward(p_score, n_score)
		return score.cpu().data.numpy()

def get_softplus_loss_hpo_config() -> dict[str, dict[str, Any]]:

	"""返回 :py:class:`SoftplusLoss` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'loss': {
				'value': 'SoftplusLoss'
			},
			'adv_temperature': {
				'values': [1.0, 3.0, 6.0]
			}
		}
	
	:returns: :py:class:`SoftplusLoss` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'loss': {
			'value': 'SoftplusLoss'
		},
		'adv_temperature': {
			'values': [1.0, 3.0, 6.0]
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/module/loss/CompGCNLoss.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 19, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 23, 2024
#
# 该脚本定义了 CompGCNLoss 类.

"""
CompGCNLoss - 损失函数类，CompGCN 原论文中应用这种损失函数完成模型学习。
"""

import torch
from .Loss import Loss
from typing import Any
from ..model import CompGCN

class CompGCNLoss(Loss):
    
    """
    ``CompGCN`` :cite:`CompGCN` 原论文中应用这种损失函数完成模型训练。
    
    .. Note:: :py:meth:`forward` 中的正样本评分函数的得分应大于负样本评分函数的得分。
    
    例子::

        from unike.module.loss import CompGCNLoss
        from unike.module.strategy import CompGCNSampling

        # define the loss function
        model = CompGCNSampling(
            model = compgcn,
            loss = CompGCNLoss(model = compgcn),
            ent_tol = dataloader.get_ent_tol()
        )
    """

    def __init__(
        self,
        model: CompGCN):

        """创建 CompGCNLoss 对象。
        
        :param model: 模型
        :type model: CompGCN
		"""

        super(CompGCNLoss, self).__init__()

        #: 模型
        self.model: CompGCN = model
        #: 损失函数
        self.loss: torch.nn.BCELoss = torch.nn.BCELoss()

    def forward(
        self,
        pred: torch.Tensor,
        label: torch.Tensor) -> torch.Tensor:

        """计算 CompGCNLoss 损失函数。定义每次调用时执行的计算。:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
        
        :param pred: 模型的得分。
        :type pred: torch.Tensor
        :param labels: 标签
        :type labels: torch.Tensor
        :returns: 损失值
        :rtype: torch.Tensor
		"""
        
        loss = self.loss(pred, label)
        return loss

def get_compgcn_loss_hpo_config() -> dict[str, dict[str, Any]]:

	"""返回 :py:class:`CompGCNLoss` 的默认超参数优化配置。
	
	默认配置为::
	
	    parameters_dict = {
	    	'loss': {
	    		'value': 'CompGCNLoss'
	    	}
	    }
	
	:returns: :py:class:`CompGCNLoss` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'loss': {
			'value': 'CompGCNLoss'
		}
	}
		
	return parameters_dict


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .version import __version__

import sys
from loguru import logger

logger.remove()
logger.add(sys.stdout, level="INFO", format="{time:YYYY-MM-DD HH:mm:ss} | {level} | {message}")



# coding:utf-8
#
# unike/data/BernSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 30, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 25, 2024
#
# 该脚本定义了 BernSampler 类.

"""
BernSampler - 平移模型和语义匹配模型的训练集数据采样器。
"""

import torch
import typing
import random
import collections
import numpy as np
from .TradSampler import TradSampler
from typing_extensions import override

class BernSampler(TradSampler):

    """
    平移模型和语义匹配模型的训练集 Bern 数据采样器（伯努利分布），如果想获得更详细的信息请访问 :ref:`TransH <transh>`。
    """

    def __init__(
        self,
        in_path: str = "./",
        ent_file: str = "entity2id.txt",
        rel_file: str = "relation2id.txt",
        train_file: str = "train2id.txt",
        batch_size: int | None = None,
        neg_ent: int = 1):

        """创建 BernSampler 对象。

        :param in_path: 数据集目录
        :type in_path: str
        :param ent_file: entity2id.txt
        :type ent_file: str
        :param rel_file: relation2id.txt
        :type rel_file: str
        :param train_file: train2id.txt
        :type train_file: str
        :param batch_size: batch size 在该采样器中不起作用，只是占位符。
        :type batch_size: int | None
        :param neg_ent: 对于每一个正三元组, 构建的负三元组的个数, 替换 entity
        :type neg_ent: int
        """

        super().__init__(
            in_path=in_path,
            ent_file=ent_file,
            rel_file=rel_file,
            train_file=train_file,
            batch_size = batch_size,
            neg_ent = neg_ent
        )
        
        self.tph, self.hpt = self.get_tph_hpt()

    def get_tph_hpt(self) -> tuple[collections.defaultdict[float], collections.defaultdict[float]]:

        """计算 tph 和 hpt。
        
        :returns: tph 和 hpt
        :rtype: tuple[collections.defaultdict[float], collections.defaultdict[float]]
        """

        h_of_r = collections.defaultdict(set)
        t_of_r = collections.defaultdict(set)
        freq_rel = collections.defaultdict(float)
        tph = collections.defaultdict(float)
        hpt = collections.defaultdict(float)
        for h, r, t in self.train_triples:
            freq_rel[r] += 1.0
            h_of_r[r].add(h)
            t_of_r[r].add(t)
        for r in h_of_r:
            tph[r] = freq_rel[r] / len(h_of_r[r])
            hpt[r] = freq_rel[r] / len(t_of_r[r])
        return tph, hpt

    @override
    def sampling(
        self,
        pos_triples: list[tuple[int, int, int]]) -> dict[str, typing.Union[str, torch.Tensor]]:

        """平移模型和语义匹配模型的训练集 bern 的数据采样函数（伯努利分布）。
        
        :param pos_triples: 知识图谱中的正确三元组
        :type pos_triples: list[tuple[int, int, int]]
        :returns: 平移模型和语义匹配模型的训练数据
        :rtype: dict[str, typing.Union[str, torch.Tensor]]
        """

        batch_data = {}
        neg_ent_sample = []

        batch_data['mode'] = 'bern'
        for h, r, t in pos_triples:
            neg_ent = self.__normal_batch(h, r, t, self.neg_ent)
            neg_ent_sample += neg_ent
        
        batch_data["positive_sample"] = torch.LongTensor(np.array(pos_triples))
        batch_data["negative_sample"] = torch.LongTensor(np.array(neg_ent_sample))

        return batch_data

    def __normal_batch(
        self,
        h: int,
        r: int,
        t: int,
        neg_size: int) -> list[tuple[int, int, int]]:

        """Bern 负采样函数
        
        :param h: 头实体
        :type h: int
        :param r: 关系
        :type r: int
        :param t: 尾实体
        :type t: int
        :param neg_size: 负三元组个数
        :type neg_size: int
        :returns: 负三元组中的头实体列表
        :rtype: list[tuple[int, int, int]]
        """

        neg_size_h = 0
        neg_size_t = 0
        prob = self.hpt[r] / (self.hpt[r] + self.tph[r])
        for _ in range(neg_size):
            if random.random() < prob:
                neg_size_t += 1
            else:
                neg_size_h += 1

        res = []

        neg_list_h = []
        neg_cur_size = 0
        while neg_cur_size < neg_size_h:
            neg_tmp_h = self.corrupt_head(t, r, num_max=(neg_size_h - neg_cur_size) * 2)
            neg_list_h.append(neg_tmp_h)
            neg_cur_size += len(neg_tmp_h)
        if neg_list_h != []:
            neg_list_h = np.concatenate(neg_list_h)
        
        for hh in neg_list_h[:neg_size_h]:
            res.append((hh, r, t))
        
        neg_list_t = []
        neg_cur_size = 0
        while neg_cur_size < neg_size_t:
            neg_tmp_t = self.corrupt_tail(h, r, num_max=(neg_size_t - neg_cur_size) * 2)
            neg_list_t.append(neg_tmp_t)
            neg_cur_size += len(neg_tmp_t)
        if neg_list_t != []:
            neg_list_t = np.concatenate(neg_list_t)
        
        for tt in neg_list_t[:neg_size_t]:
            res.append((h, r, tt))

        return res


# coding:utf-8
#
# unike/module/BaseModule.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 4, 2023
# 
# 该头文件定义了 BaseModule.

"""BaseModule - 所有模块的基类"""

import os
import json
import torch
import torch.nn as nn
import numpy as np
from typing import Any

class BaseModule(nn.Module):

	"""继承自 :py:class:`torch.nn.Module`，并且封装了一些常用功能，如加载和保存模型。"""

	def __init__(self):

		"""创建 BaseModule 对象。"""

		super(BaseModule, self).__init__()

		#: 常数 0
		self.zero_const: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor([0]))
		self.zero_const.requires_grad = False

		#: 常数 pi
		self.pi_const: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor([3.14159265358979323846]))
		self.pi_const.requires_grad = False

	def load_checkpoint(self, path: str):
		
		"""加载模型权重。

		:param path: 模型保存的路径
		:type path: str
		"""

		self.load_state_dict(torch.load(os.path.join(path)))
		self.eval()

	def save_checkpoint(self, path: str):

		"""保存模型权重。

		:param path: 模型保存的路径
		:type path: str
		"""

		if not os.path.exists(os.path.split(path)[0]):
			os.makedirs(os.path.split(path)[0], exist_ok=True)
		torch.save(self.state_dict(), path)

	def get_parameters(
		self,
		mode: str = "numpy",
		param_dict: dict[str, Any] | None = None
		) -> dict[str, np.ndarray] | dict[str, list] | dict[str, torch.Tensor]:

		"""获得模型权重。

		:param mode: 模型保存的格式，可以选择 ``numpy`` 、 ``list`` 和 ``Tensor`` 。
		:type path: str
		:param param_dict: 可以选择从哪里获得模型权重。
		:type param_dict: dict[str, typing.Any] | None
		:returns: 模型权重字典。
		:rtype: dict[str, numpy.ndarray] | dict[str, list] | dict[str, torch.Tensor]
		"""

		all_param_dict = self.state_dict()
		if param_dict == None:
			param_dict = all_param_dict.keys()
		res = {}
		for param in param_dict:
			if mode == "numpy":
				res[param] = all_param_dict[param].cpu().numpy()
			elif mode == "list":
				res[param] = all_param_dict[param].cpu().numpy().tolist()
			else:
				res[param] = all_param_dict[param]
		return res

	def set_parameters(self, parameters: dict[str, Any]):
		
		"""加载模型权重。

		:param parameters: 模型权重字典。
		:type parameters: dict[str, typing.Any]
		"""

		for i in parameters:
			parameters[i] = torch.Tensor(parameters[i])
		self.load_state_dict(parameters, strict = False)
		self.eval()

	def load_parameters(self, path: str):

		"""加载模型权重。

		:param path: 模型保存的路径
		:type path: str
		"""

		f = open(path, "r")
		parameters = json.loads(f.read())
		f.close()
		for i in parameters:
			parameters[i] = torch.Tensor(parameters[i])
		self.load_state_dict(parameters, strict = False)
		self.eval()

	def save_parameters(self, path: str):

		"""用 json 格式保存模型权重。

		:param path: 模型保存的路径
		:type path: str
		"""

		if not os.path.exists(os.path.split(path)[0]):
			os.makedirs(os.path.split(path)[0], exist_ok=True)
		f = open(path, "w")
		f.write(json.dumps(self.get_parameters("list")))
		f.close()


# coding:utf-8
#
# unike/module/strategy/CompGCNSampling.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 20, 2023
#
# 该脚本定义了 CompGCN 模型的训练策略.

"""
CompGCNSampling - 训练策略类，包含损失函数。
"""

import dgl
import torch
import typing
from ..loss import Loss
from ..model import CompGCN
from .Strategy import Strategy

class CompGCNSampling(Strategy):

	"""
	将模型和损失函数封装到一起，方便模型训练，用于 ``CompGCN`` :cite:`CompGCN`。

	例子::

		from unike.module.model import CompGCN
		from unike.module.loss import CompGCNLoss
		from unike.module.strategy import CompGCNSampling
		from unike.config import Trainer, GraphTester
		
		# define the model
		compgcn = CompGCN(
			ent_tol = dataloader.train_sampler.ent_tol,
			rel_tol = dataloader.train_sampler.rel_tol,
			dim = 100
		)
		
		# define the loss function
		model = CompGCNSampling(
			model = compgcn,
			loss = CompGCNLoss(model = compgcn),
			ent_tol = dataloader.train_sampler.ent_tol
		)
		
		# test the model
		tester = GraphTester(model = compgcn, data_loader = dataloader, use_gpu = True, device = 'cuda:0', prediction = "tail")
		
		# train the model
		trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(),
			epochs = 2000, lr = 0.0001, use_gpu = True, device = 'cuda:0',
			tester = tester, test = True, valid_interval = 50, log_interval = 50,
			save_interval = 50, save_path = '../../checkpoint/compgcn.pth'
		)
		trainer.run()
	"""

	def __init__(
		self,
		model: CompGCN = None,
		loss: Loss = None,
		smoothing: float = 0.1,
		ent_tol: int = None):
		
		"""创建 CompGCNSampling 对象。
		
		:param model: CompGCN 模型
		:type model: :py:class:`unike.module.model.CompGCN`
		:param loss: 损失函数。
		:type loss: :py:class:`unike.module.loss.Loss`
		:param smoothing: smoothing
		:type smoothing: float
		:param ent_tol: 实体个数
		:type ent_tol: int
		"""

		super(CompGCNSampling, self).__init__()

		#: CompGCN 模型，即 :py:class:`unike.module.model.CompGCN`
		self.model: CompGCN = model
		#: 损失函数，即 :py:class:`unike.module.loss.Loss`
		self.loss: Loss = loss
		#: smoothing
		self.smoothing: float = smoothing
		#: 实体个数
		self.ent_tol: int = ent_tol

	def forward(
		self,
		data: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]) -> torch.Tensor:
		
		"""计算最后的损失值。定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param data: 数据
		:type data: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]
		:returns: 损失值
		:rtype: torch.Tensor
		"""
		
		graph    = data["graph"]
		relation = data['relation']
		norm     = data['norm']
		sample   = data["sample"]
		label    = data["label"]
		score = self.model(graph, relation, norm, sample)
		label = (1.0 - self.smoothing) * label + (1.0 / self.ent_tol)
		loss  = self.loss(score,  label)
		return loss

def get_compgcn_sampling_hpo_config() -> dict[str, dict[str, typing.Any]]:
	
	"""返回 :py:class:`CompGCNSampling` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'strategy': {
				'value': 'CompGCNSampling'
			},
			'smoothing': {
				'value': 0.1
			}
		}
	
	:returns: :py:class:`CompGCNSampling` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""
	
	parameters_dict = {
		'strategy': {
			'value': 'CompGCNSampling'
		},
		'smoothing': {
			'value': 0.1
		}
	}
	
	return parameters_dict


# coding:utf-8
#
# unike/data/TradSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 28, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 25, 2024
#
# 为 KGReader 增加构建负三元组的函数，用于平移模型和语义匹配模型.

"""
TradSampler - 为 KGReader 增加构建负三元组的函数，用于平移模型和语义匹配模型。
"""

import torch
import typing
import numpy as np
from .KGReader import KGReader

class TradSampler(KGReader):
    
    """
    平移模型和语义匹配模型的采样器的基类。
    """
    
    def __init__(
        self,
        in_path: str = "./",
        ent_file: str = "entity2id.txt",
        rel_file: str = "relation2id.txt",
        train_file: str = "train2id.txt",
        batch_size: int | None = None,
        neg_ent: int = 1):

        """创建 TradSampler 对象。

        :param in_path: 数据集目录
        :type in_path: str
        :param ent_file: entity2id.txt
        :type ent_file: str
        :param rel_file: relation2id.txt
        :type rel_file: str
        :param train_file: train2id.txt
        :type train_file: str
        :param batch_size: batch size 在该采样器中不起作用，只是占位符。
        :type batch_size: int | None
        :param neg_ent: 对于每一个正三元组, 构建的负三元组的个数, 替换 entity
        :type neg_ent: int
        """
        
        super().__init__(
            in_path=in_path,
            ent_file=ent_file,
            rel_file=rel_file,
            train_file=train_file
        )

        #: batch size
        self.batch_size: int = batch_size
        #: 对于每一个正三元组, 构建的负三元组的个数, 替换 entity (head + tail)
        self.neg_ent: int = neg_ent

        self.get_hr2t_rt2h_from_train()

    def sampling(
        self,
        pos_triples: list[tuple[int, int, int]]) -> dict[str, typing.Union[str, torch.Tensor]]:
        
        """平移模型和语义匹配模型的训练集数据采样函数。该方法未实现，子类必须重写该方法，否则抛出 :py:class:`NotImplementedError` 错误。
        
        :param pos_triples: 知识图谱中的正确三元组
        :type pos_triples: list[tuple[int, int, int]]
        :returns: 平移模型和语义匹配模型的训练数据
        :rtype: dict[str, typing.Union[str, torch.Tensor]]
        """
        
        raise NotImplementedError

    def corrupt_head(
        self,
        t: int,
        r: int,
        num_max: int = 1) -> np.ndarray:

        """替换头实体构建负三元组。

        :param t: 尾实体
        :type t: int
        :param r: 关系
        :type r: int
        :param num_max: 一次负采样的个数
        :type num_max: int
        :returns: 负三元组的头实体列表
        :rtype: numpy.ndarray
        """
        
        tmp = torch.randint(low=0, high=self.ent_tol, size=(num_max,)).numpy()
        mask = np.in1d(tmp, self.rt2h_train[(r, t)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg
        
    def corrupt_tail(
        self,
        h: int,
        r: int,
        num_max: int = 1) -> np.ndarray:
        
        """替换尾实体构建负三元组。

        :param h: 头实体
        :type h: int
        :param r: 关系
        :type r: int
        :param num_max: 一次负采样的个数
        :type num_max: int
        :returns: 负三元组的尾实体列表
        :rtype: numpy.ndarray
        """
        
        tmp = torch.randint(low=0, high=self.ent_tol, size=(num_max,)).numpy()
        mask = np.in1d(tmp, self.hr2t_train[(h, r)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg


# coding:utf-8
#
# unike/data/CompGCNTestSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 19, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 19, 2023
#
# 该脚本定义了 CompGCNTestSampler 类.

"""
CompGCNTestSampler - CompGCN 的测试数据采样器。
"""

from .CompGCNSampler import CompGCNSampler
from .RGCNTestSampler import RGCNTestSampler

class CompGCNTestSampler(RGCNTestSampler):

    """``CompGCN`` :cite:`CompGCN` 的测试数据采样器。

    例子::

        from unike.data import RGCNTestSampler, CompGCNTestSampler
        from torch.utils.data import DataLoader

        #: 测试数据采样器
        test_sampler: typing.Union[typing.Type[RGCNTestSampler], typing.Type[CompGCNTestSampler]] = test_sampler(
            sampler=train_sampler,
            valid_file=valid_file,
            test_file=test_file,
        )
    
        #: 验证集三元组
        data_val: list[tuple[int, int, int]] = test_sampler.get_valid()
        #: 测试集三元组
        data_test: list[tuple[int, int, int]] = test_sampler.get_test()

        val_dataloader = DataLoader(
            data_val,
            shuffle=False,
            batch_size=test_batch_size,
            num_workers=num_workers,
            pin_memory=True,
            collate_fn=test_sampler.sampling,
        )

        test_dataloader = DataLoader(
            data_test,
            shuffle=False,
            batch_size=test_batch_size,
            num_workers=num_workers,
            pin_memory=True,
            collate_fn=test_sampler.sampling,
        )
    """

    def __init__(
        self,
        sampler: CompGCNSampler,
        valid_file: str = "valid2id.txt",
        test_file: str = "test2id.txt",
        type_constrain: bool = True):

        """创建 CompGCNTestSampler 对象。

        :param sampler: 训练数据采样器。
        :type sampler: CompGCNSampler
        :param valid_file: valid2id.txt
        :type valid_file: str
        :param test_file: test2id.txt
        :type test_file: str
        :param type_constrain: 是否报告 type_constrain.txt 限制的测试结果
        :type type_constrain: bool
        """

        super().__init__(
            sampler=sampler,
            valid_file = valid_file,
            test_file = test_file,
            type_constrain = type_constrain
        )
        
        #: 幂
        self.power: float = -0.5


# coding:utf-8
#
# unike/utils/__init__.py
# 
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on July 6, 2023
# 
# 该头文件定义了 utils 接口.

"""工具类。"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .Timer import Timer
from .WandbLogger import WandbLogger
from .tools import import_class, construct_type_constrain
from .EarlyStopping import EarlyStopping
from .Link import Link

__all__ = [
	'Timer',
	'WandbLogger',
	'import_class',
	'construct_type_constrain',
	'EarlyStopping',
	'Link',
]


# coding:utf-8
#
# unike/data/RevSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 15, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2024
#
# 为 KGReader 增加相反关系，用于图神经网络模型.

"""
RevSampler - 为 KGReader 增加相反关系，用于图神经网络模型。
"""

import os
from .KGReader import KGReader

class RevSampler(KGReader):
    
    """在训练集中增加相反关系.

    对于每一个三元组 (h, r, t)，生成相反关系三元组 (t, r`, h): r` = r + rel_tol。
    """
    
    def __init__(
        self,
        in_path: str = "./",
        ent_file: str = "entity2id.txt",
        rel_file: str = "relation2id.txt",
        train_file: str = "train2id.txt"):

        """创建 RevSampler 对象。

        :param in_path: 数据集目录
        :type in_path: str
        :param ent_file: entity2id.txt
        :type ent_file: str
        :param rel_file: relation2id.txt
        :type rel_file: str
        :param train_file: train2id.txt
        :type train_file: str
        """
        
        super().__init__(
            in_path=in_path,
            ent_file=ent_file,
            rel_file=rel_file,
            train_file=train_file
        )

        self.add_reverse_relation()
        self.add_train_reverse_triples()
        self.get_hr2t_rt2h_from_train()
        
    def add_reverse_relation(self):

        """增加相反关系：r` = r + rel_tol"""
        
        with open(os.path.join(self.in_path, self.rel_file)) as f:
            f.readline()
            for line in f:
                relation, rid = line.strip().split("\t")
                self.rel2id[relation + "_reverse"] = int(rid) + self.rel_tol
                self.id2rel[int(rid) + self.rel_tol] = relation + "_reverse"
        self.rel_tol = len(self.rel2id)
        
    def add_train_reverse_triples(self):

        """对于每一个三元组 (h, r, t)，生成相反关系三元组 (t, r`, h): r` = r + rel_tol。"""

        tol = int(self.rel_tol / 2)
        
        with open(os.path.join(self.in_path, self.train_file)) as f:
            f.readline()
            for line in f:
                h, t, r = line.strip().split()
                self.train_triples.append(
                    (int(t), int(r) + tol, int(h))
                )


# coding:utf-8
#
# unike/module/strategy/NegativeSampling.py
#
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 9, 2024
#
# 该脚本定义了平移模型和语义匹配模型的训练策略.

"""
NegativeSampling - 训练策略类，包含损失函数。
"""

import torch
import typing
from ..loss import Loss
from ..model import Model
from .Strategy import Strategy

class NegativeSampling(Strategy):

	"""
	将模型和损失函数封装到一起，方便模型训练。

	例子::

		from unike.module.model import TransE
		from unike.module.loss import MarginLoss
		from unike.module.strategy import NegativeSampling
		
		# define the model
		transe = TransE(
			ent_tol = dataloader.get_ent_tol(),
			rel_tol = dataloader.get_rel_tol(),
			dim = 50, 
			p_norm = 1, 
			norm_flag = True
		)
		
		# define the loss function
		model = NegativeSampling(
			model = transe, 
			loss = MarginLoss(margin = 1.0),
			regul_rate = 0.01
		)
	"""

	def __init__(
		self,
		model: Model = None,
		loss: Loss = None,
		regul_rate: float = 0.0,
		l3_regul_rate: float = 0.0):
		
		"""创建 NegativeSampling 对象。

		:param model: KGE 模型
		:type model: :py:class:`unike.module.model.Model`
		:param loss: 损失函数。
		:type loss: :py:class:`unike.module.loss.Loss`
		:param regul_rate: 权重衰减系数
		:type regul_rate: float
		:param l3_regul_rate: l3 正则化系数
		:type l3_regul_rate: float
		"""

		super(NegativeSampling, self).__init__()
		#: KGE 模型，即 :py:class:`unike.module.model.Model`
		self.model: Model = model
		#: 损失函数，即 :py:class:`unike.module.loss.Loss`
		self.loss: Loss = loss
		#: 权重衰减系数
		self.regul_rate: float = regul_rate
		#: l3 正则化系数
		self.l3_regul_rate: float = l3_regul_rate

	def forward(self, data: dict[str, typing.Union[torch.Tensor, str]]) -> torch.Tensor:
		
		"""计算最后的损失值。定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param data: 数据
		:type data: dict[str, typing.Union[torch.Tensor, str]]
		:returns: 损失值
		:rtype: torch.Tensor
		"""
		
		pos_sample = data["positive_sample"]
		neg_sample = data["negative_sample"]
		mode = data["mode"]
		pos_score = self.model(pos_sample)
		if mode == "bern":
			neg_score = self.model(neg_sample)
			neg_score = neg_score.view(pos_score.shape[0], -1)
		else:
			neg_score = self.model(pos_sample, neg_sample, mode)
		loss_res = self.loss(pos_score, neg_score)
		if self.regul_rate != 0:
			loss_res += self.regul_rate * self.model.regularization(data)
		if self.l3_regul_rate != 0:
			loss_res += self.l3_regul_rate * self.model.l3_regularization()
		return loss_res

def get_negative_sampling_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`NegativeSampling` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'strategy': {
				'value': 'NegativeSampling'
			},
			'regul_rate': {
				'value': 0.0
			},
			'l3_regul_rate': {
				'value': 0.0
			}
		}
	
	:returns: :py:class:`NegativeSampling` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'strategy': {
			'value': 'NegativeSampling'
		},
		'regul_rate': {
			'value': 0.0
		},
		'l3_regul_rate': {
			'value': 0.0
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/config/HPOTrainer.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 2, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 6, 2024
#
# 该脚本定义了并行训练循环函数.

"""
hpo_train - 超参数优化训练循环函数。
"""

import wandb
import typing
from ..utils import import_class
from ..module.model import TransE
from ..module.loss import MarginLoss
from ..module.strategy import NegativeSampling
from ..config import Trainer, Tester
from ..data import KGEDataLoader
from ..utils import WandbLogger
from loguru import logger
from typing import Literal, Any, Optional
import os


def set_hpo_config(
	method: str = 'bayes',
	sweep_name: str = 'unike_hpo',
	metric_name: str = 'val/hits@10',
	metric_goal: Literal['maximize', 'minimize'] = 'maximize',
	data_loader_config: dict[str, dict[str, typing.Any]] = {},
	kge_config: dict[str, dict[str, typing.Any]] = {},
	loss_config: dict[str, dict[str, typing.Any]] = {},
	strategy_config: dict[str, dict[str, typing.Any]] = {},
	tester_config: dict[str, dict[str, typing.Any]] = {},
	trainer_config: dict[str, dict[str, typing.Any]] = {}) -> dict[str, Any]:

	"""设置超参数优化范围。
	
	:param method: 超参数优化的方法，``grid`` 或 ``random`` 或 ``bayes``
	:type param: str
	:param sweep_name: 超参数优化 sweep 的名字
	:type sweep_name: str
	:param metric_name: 超参数优化的指标名字
	:type metric_name: str
	:param metric_goal: 超参数优化的指标目标，``maximize`` 或 ``minimize``
	:type metric_goal: str
	:param data_loader_config: :py:class:`unike.data.KGEDataLoader` 的超参数优化配置
	:type data_loader_config: dict
	:param kge_config: :py:class:`unike.module.model.Model` 的超参数优化配置
	:type kge_config: dict
	:param loss_config: :py:class:`unike.module.loss.Loss` 的超参数优化配置
	:type loss_config: dict
	:param strategy_config: :py:class:`unike.module.strategy.Strategy` 的超参数优化配置
	:type strategy_config: dict
	:param tester_config: :py:class:`unike.config.Tester` 的超参数优化配置
	:type tester_config: dict
	:param trainer_config: :py:class:`unike.config.Trainer` 的超参数优化配置
	:type trainer_config: dict
	:returns: 超参数优化范围
	:rtype: dict
	"""

	sweep_config: dict[str, Any] = {
		'method': method,
		'name': sweep_name
	}

	metric: dict[str, str] = {
		'name': metric_name,
		'goal': metric_goal
	}

	parameters_dict: dict[str, dict[str, typing.Any]] | None = {}
	parameters_dict.update(data_loader_config)
	parameters_dict.update(kge_config)
	parameters_dict.update(loss_config)
	parameters_dict.update(strategy_config)
	parameters_dict.update(tester_config)
	parameters_dict.update(trainer_config)

	sweep_config['metric'] = metric
	sweep_config['parameters'] = parameters_dict

	return sweep_config

def set_hpo_hits(
    new_hits: list[int] = [1, 3, 10]):
	
	"""设置 Hits 指标。
	
	:param new_hits: 准备报告的指标 Hit@N 的列表，默认为 [1, 3, 10], 表示报告 Hits@1, Hits@3, Hits@10
	:type new_hits: list[int]
    """
	
	tmp = Tester.hits
	Tester.hits = new_hits
	logger.info(f"Hits@N 指标由 {tmp} 变为 {Tester.hits}")

def start_hpo_train(
	config: dict[str, Any],
	project: str = "unike-sweeps",
	count: int = 2,
	prior_runs: Optional[list[str]] = None,
 	resume_sweep_id: str | None = None) -> None:

	"""开启超参数优化。
	
	:param config: wandb 的超参数优化配置。
	:type config: dict
	:param project: 项目名
	:type param: str
	:param count: 进行几次尝试。
	:type count: int
	:param resume_sweep_id: 恢复的超参数 sweep id
	:type resume_sweep_id: str | None
	"""
	wandb.login()
 
	if resume_sweep_id:
		api = wandb.Api()
		sweep_id = f'{api.default_entity}/{project}/{resume_sweep_id}'
	else:
		sweep_id = wandb.sweep(config, project=project, prior_runs=prior_runs)

	wandb.agent(sweep_id, hpo_train, count=count)

def hpo_train(config: dict[str, Any] | None = None):

	"""超参数优化训练循环函数。
	
	:param config: wandb 的项目配置如超参数。
	:type config: dict[str, Any]
	"""
	
	with wandb.init(config = config):
		
		config_ = wandb.config

		# dataloader for training
		dataloader_class: type[KGEDataLoader] = import_class(f"unike.data.{config_.dataloader}")
		dataloader = dataloader_class(
			in_path = config_.in_path,
			ent_file = config_.ent_file,
			rel_file = config_.rel_file,
			train_file = config_.train_file,
			valid_file = config_.valid_file,
			test_file = config_.test_file,
			batch_size = config_.batch_size,
			neg_ent = config_.neg_ent,
			test = True,
			test_batch_size = config_.test_batch_size,
			type_constrain = config_.type_constrain,
			num_workers = config_.num_workers,
			train_sampler = import_class(f"unike.data.{config_.train_sampler}"),
			test_sampler = import_class(f"unike.data.{config_.test_sampler}")
		)

		# define the model
		if config_.model in ["TransE", "TransH"]:
			model_class = import_class(f"unike.module.model.{config_.model}")
			kge_model = model_class(
			    ent_tol = dataloader.get_ent_tol(),
			    rel_tol = dataloader.get_rel_tol(),
			    dim = config_.dim,
			    p_norm = config_.p_norm,
			    norm_flag = config_.norm_flag
			)
		elif config_.model == "TransR":
			model_class = import_class(f"unike.module.model.{config_.model}")
			transe = TransE(
				ent_tol = dataloader.get_ent_tol(),
				rel_tol = dataloader.get_rel_tol(),
				dim = config_.dim,
				p_norm = config_.p_norm,
				norm_flag = config_.norm_flag
			)
			kge_model = model_class(
				ent_tol = dataloader.get_ent_tol(),
				rel_tol = dataloader.get_rel_tol(),
				dim_e = config_.dim,
				dim_r = config_.dim,
				p_norm = config_.p_norm,
				norm_flag = config_.norm_flag,
				rand_init = config_.rand_init)
			model_e = NegativeSampling(
				model = transe,
				loss = MarginLoss(margin = config_.margin_e)
			)
			trainer_e = Trainer(
				model = model_e,
				data_loader = dataloader.train_dataloader(),
				epochs = 1,
				lr = config_.lr_e,
				opt_method = config_.opt_method_e,
				use_gpu = config_.use_gpu_trainer,
				device = config_.device_trainer
			)
			trainer_e.run()
			parameters = transe.get_parameters()
			transe.save_parameters("./transr_transe.json")
			kge_model.set_parameters(parameters)
		elif config_.model == "TransD":
			model_class = import_class(f"unike.module.model.{config_.model}")
			kge_model = model_class(
				ent_tol = dataloader.get_ent_tol(),
				rel_tol = dataloader.get_rel_tol(),
				dim_e = config_.dim_e,
				dim_r = config_.dim_r,
				p_norm = config_.p_norm,
				norm_flag = config_.norm_flag)
		elif config_.model == "RotatE":
			model_class = import_class(f"unike.module.model.{config_.model}")
			kge_model = model_class(
				ent_tol = dataloader.get_ent_tol(),
				rel_tol = dataloader.get_rel_tol(),
				dim = config_.dim,
				margin = config_.margin,
				epsilon = config_.epsilon)
		elif config_.model in ["RESCAL", "DistMult", "HolE", "ComplEx", "Analogy", "SimplE"]:
			model_class = import_class(f"unike.module.model.{config_.model}")
			kge_model = model_class(
			    ent_tol = dataloader.get_ent_tol(),
			    rel_tol = dataloader.get_rel_tol(),
			    dim = config_.dim)
		elif config_.model == "RGCN":
			model_class = import_class(f"unike.module.model.{config_.model}")
			kge_model = model_class(
				ent_tol = dataloader.get_ent_tol(),
				rel_tol = dataloader.get_rel_tol(),
				dim = config_.dim,
				num_layers = config_.num_layers)
		elif config_.model == "CompGCN":
			model_class = import_class(f"unike.module.model.{config_.model}")
			kge_model = model_class(
				ent_tol = dataloader.get_ent_tol(),
				rel_tol = dataloader.get_rel_tol(),
				dim = config_.dim,
				opn = config_.opn,
				fet_drop = config_.fet_drop,
				hid_drop = config_.hid_drop,
				margin = config_.margin,
				decoder_model = config_.decoder_model)
		else:
			import sys
			sys.path.append(config_.model_class_path)
			model_class = import_class(f"{config_.model_class}")
			params_config = {key: value for key, value in config_.items() if key in model_class.__init__.__code__.co_varnames}
			kge_model = model_class(
       			ent_tol = dataloader.get_ent_tol(),
			    rel_tol = dataloader.get_rel_tol(),
       			**params_config
          	)

		# define the loss function
		loss_class = import_class(f"unike.module.loss.{config_.loss}")
		if config_.loss == 'MarginLoss':
			loss = loss_class(
				adv_temperature = config_.adv_temperature,
				margin = config_.margin
			)
		elif config_.loss in ['SigmoidLoss', 'SoftplusLoss']:
			loss = loss_class(adv_temperature = config_.adv_temperature)
		elif config_.loss == 'RGCNLoss':
			loss = loss_class(
				model = kge_model,
				regularization = config_.regularization
			)
		elif config_.loss == 'CompGCNLoss':
			loss = loss_class(model = kge_model)
		
		# define the strategy
		strategy_class = import_class(f"unike.module.strategy.{config_.strategy}")
		if config_.strategy == 'NegativeSampling':
			model = strategy_class(
				model = kge_model,
				loss = loss,
				regul_rate = config_.regul_rate,
				l3_regul_rate = config_.l3_regul_rate
			)
		elif config_.strategy == 'RGCNSampling':
			model = strategy_class(
				model = kge_model,
				loss = loss
			)
		elif config_.strategy == 'CompGCNSampling':
			model = strategy_class(
				model = kge_model,
				loss = loss,
				smoothing = config_.smoothing,
				ent_tol = dataloader.train_sampler.ent_tol
			)

		# test the model
		tester_class: type[Tester] = import_class(f"unike.config.{config_.tester}")
		tester = tester_class(
			model = kge_model,
			data_loader = dataloader,
			prediction = config_.prediction,
			use_tqdm = config_.use_tqdm,
			use_gpu = config_.use_gpu_tester,
			device = config_.device_tester
		)
  
		wandb_logger = WandbLogger(endpoint='wandb')
		wandb_logger.logger = wandb

		# # train the model
		trainer_class: type[Trainer] = import_class(f"unike.config.{config_.trainer}")
		trainer = trainer_class(
			model = model,
			data_loader = dataloader.train_dataloader(),
			epochs = config_.epochs,
			lr = config_.lr,
			opt_method = config_.opt_method,
			use_gpu = config_.use_gpu_trainer,
			device = config_.device_trainer,
			tester = tester,
			test = True,
			valid_interval = config_.valid_interval,
			log_interval = config_.log_interval,
			save_path = config_.save_path,
			use_early_stopping = config_.use_early_stopping,
			metric = config_.metric,
			patience = config_.patience,
			delta = config_.delta,
			wandb_logger = wandb_logger
		)
		trainer.run()



# coding:utf-8
#
# unike/utils/tools.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 3, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 25, 2024
#
# 该脚本定义了 WandbLogger 类.

import importlib

def import_class(module_and_class_name: str) -> type:

    """从模块中导入类。
    
    :param module_and_class_name: 模块和类名，如 **unike.module.model.TransE** 。
    :type module_and_class_name: str
    :returns: 类名
    :rtype: type
    """

    module_name, class_name = module_and_class_name.rsplit(".", 1)
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name)
    return class_

def construct_type_constrain(
    in_path: str = "./",
    train_file: str = "train2id.txt",
    valid_file: str = "valid2id.txt",
    test_file: str = "test2id.txt"
    ):

    """构建 type_constrain.txt 文件
    
    type_constrain.txt: 类型约束文件, 第一行是关系的个数
    
    下面的行是每个关系的类型限制 (训练集、验证集、测试集中每个关系存在的 head 和 tail 的类型)
    
    每个关系有两行：
    
    第一行：**rel_id** **heads_num** **head1** **head2** ...
    
    第二行: **rel_id** **tails_num** **tail1** **tail2** ...

    如 benchmarks/FB15K 的 id 为 1200 的关系，它有 4 种类型头实体（3123，1034，58 和 5733）和 4 种类型的尾实体（12123，4388，11087 和 11088）。
    
    1200	4	3123	1034	58	5733

    1200	4	12123	4388	11087	11088
    
    :param in_path: 数据集目录
    :type in_path: str
    :param train_file: train2id.txt
    :type train_file: str
    :param valid_file: valid2id.txt
    :type valid_file: str
    :param test_file: test2id.txt
    :type test_file: str
    """


    rel_head: dict = {}
    rel_tail: dict = {}
    
    train = open(in_path + train_file, "r")
    valid = open(in_path + valid_file, "r")
    test = open(in_path + test_file, "r")
    
    tot = (int)(train.readline())
    for i in range(tot):
        content = train.readline()
        h,t,r = content.strip().split()
        if not r in rel_head:
            rel_head[r] = {}
        if not r in rel_tail:
            rel_tail[r] = {}
        rel_head[r][h] = 1
        rel_tail[r][t] = 1
    
    tot = (int)(valid.readline())
    for i in range(tot):
        content = valid.readline()
        h,t,r = content.strip().split()
        if not r in rel_head:
            rel_head[r] = {}
        if not r in rel_tail:
            rel_tail[r] = {}
        rel_head[r][h] = 1
        rel_tail[r][t] = 1
        
    tot = (int)(test.readline())
    for i in range(tot):
        content = test.readline()
        h,t,r = content.strip().split()
        if not r in rel_head:
            rel_head[r] = {}
        if not r in rel_tail:
            rel_tail[r] = {}
        rel_head[r][h] = 1
        rel_tail[r][t] = 1
    
    train.close()
    valid.close()
    test.close()
    
    f = open(in_path + "type_constrain.txt", "w")
    f.write("%d\n" % (len(rel_head)))
    for i in rel_head:
        f.write("%s\t%d" % (i, len(rel_head[i])))
        for j in rel_head[i]:
            f.write("\t%s" % (j))
        f.write("\n")
        f.write("%s\t%d" % (i, len(rel_tail[i])))
        for j in rel_tail[i]:
            f.write("\t%s" % (j))
        f.write("\n")
    f.close()